import SCATPY
from pySCATMECH.model import *
from pySCATMECH.mueller import *
import numpy as np

class Local_BRDF_Model(Model):
    """
    A class for handling local BRDF Models (ones that return differential 
    scattering cross section of isolated particles
    """
    def __init__(self, *args, **kwargs):
        """
        Create and allocate a SCATMECH Local_BRDF_Model 
        
        Parameters
        ----------
        name : string 
            The name of a SCATMECH model inheriting Local_BRDF_Model

        parameters : dict of string : string
            Initializing parameters for the model.  The dict items can be
            of the form 'param' : 'value', 'param.subparam' : 'value', 
            or 'param' : dict, where the first dict item is 
            None : 'submodelname' and 
            subsequent items are parameters of submodelname.
        
        Returns
        -------
        r : Local_BRDF_Model
            The object containing a handle to the requested Local_BRDF_Model
        """
        
        self.newer = SCATPY.Get_Local_BRDF_Model
        self.freer = SCATPY.Free_Local_BRDF_Model
        if len(args)>0 and type(args[0])==str:
                self.handle = self.newer(args[0])
                args = args[1::]
        else:
            self.handle = self.newer("Bobbert_Vlieger_BRDF_Model")
        self.setParameters(*args, **kwargs)

    def __del__(self):
        SCATPY.Free_Local_BRDF_Model(self.handle)

    def MuellerDSC(self, thetai=0, thetas=0, phis=0, rotation=0, coords="psps"):
        """
        Evaluate the Mueller matrix differential scattering cross section 
        in a given geometry, and coordinate system

        Parameters
        ----------
        thetai : float
            Polar angle of incidence in radians

        thetas : float
            Polar angle of viewing in radians 

        phis: float
            Azimuthal angle of viewing in radians 
            (Note: Specular occurs when thetai = thetas, phis = 0)

        rotation : float
            Sample rotation angle in radians

        coords : string
            The coordinate system for the Mueller matrix, 
            either "psps", "xyxy", or "plane"

        Returns
        -------
        r : MuellerMatrix 
            The Mueller matrix differential scattering cross section
        """
        return MuellerMatrix(
            SCATPY.LocalDSC(self.handle, thetai, thetas, phis,
                            rotation, coords))

    def JonesDSC(self, thetai=0, thetas=0, phis=0, rotation=0, coords="psps"):
        """
        Evaluate the Mueller matrix differential scattering cross section 
        in a given geometry, and coordinate system. The result, when typecast
        to a Mueller matrix, is the Mueller matrix DSC, but maintains phase, 
        if the model has phase information.

        Parameters
        ----------
        thetai : float
            Polar angle of incidence in radians

        thetas : float
            Polar angle of viewing in radians 

        phis: float
            Azimuthal angle of viewing in radians 
            (Note: Specular occurs when thetai = thetas, phis = 0)

        rotation : float
            Sample rotation angle in radians

        coords : string
            The coordinate system for the Mueller matrix, 
            either "psps", "xyxy", or "plane"

        Returns
        -------
        r : numpy.array of complex (2x2)
            The Jones matrix differential scattering cross section
        """
        return np.array(SCATPY.LocalDSCJones(self.handle, thetai, thetas, phis,
                            rotation, coords))

    def DSC(self, thetai=0, thetas=0, phis=0, rotation=0, coords="psps",
            inc=StokesVector([1,0,0,0]),sens=StokesVector([1,0,0,0])):
        """
        Return the differential scattering cross section in a given geometry.

        Parameters
        ----------
        thetai : float
            Polar angle of incidence in radians

        thetas : float
            Polar angle of viewing in radians 

        phis : float
            Azimuthal angle of viewing in radians 
            (Note: Specular occurs when thetai = thetas, phis = 0)

        rotation : float
            Sample rotation angle in radians

        coords : string
            The coordinate system for the Mueller matrix, 
            either "psps", "xyxy", or "plane"

        inc : StokesVector
            Incident polarization as Stokes vector

        sens : StokesVector
            Polarization sensitivity of the viewer as a Stokes vector 

        Returns
        -------
        r : float
            The Mueller matrix differential scattering cross section for the 
            given incident and viewing directions and polarizations
        """
        return float((StokesVector(sens) @
                      MuellerMatrix(
                          SCATPY.LocalDSC(self.handle,
                                          thetai, thetas, phis, rotation,
                                          coords))) @ StokesVector(inc))




import SCATPY
import numpy as np
import math
import cmath

def unit(a):
    """
    Return the unit matrix associated with a vector.
    
    Parameters
    ----------
    a : np.array
        Unnormalized vector

    Returns
    -------
    unit : np.array  
           Vector such that a.dot(a) == 1.

    """
    return a/np.linalg.norm(a)

def perpto(a,b):
    """
    Return a unit vector perpendicular to two vectors.

    Parameters
    ----------
    a, b : np.array
           Two 3-vectors

    Returns
    -------
    c : np.array
        A unit vector perpendicular to both a and b.

    """
    c = np.cross(a,b)
    if np.linalg.norm(a)!=0:
        return unit(np.cross(a,b))
    else:
        c = np.cross(a,np.array([1,0,0]))
        if np.linalg.norm(a)!=0:
            return unit(np.cross(a,b))
        else:
            c = np.cross(a,np.array([0,0,1]))
            if np.linalg.norm(a)!=0:
                return unit(np.cross(a,b))
            else:
                return [1,0,0]
            
pi = math.pi # A useful symbol
deg = pi/180 # Another useful symbol
            
class MuellerMatrix(np.ndarray):
    """
    A class for handling Mueller matrices 

    Parameters
    ----------
    x : 4x4 list or array of float or int, optional
        Default is a zero Mueller matrix

    """
    def __new__(cls,*x):
        if (len(x)==0):
            m = np.array(((0.,0.,0.,0.),
                          (0.,0.,0.,0.),
                          (0.,0.,0.,0.),
                          (0.,0.,0.,0.))).view(MuellerMatrix)
            return m
        if (len(x)==16):
            m = np.asarray(x,dtype=float).view(MuellerMatrix)
            m = np.resize(m,(4,4))
            return m
        if (len(x)==4):
            for xx in x:
                if (len(xx)!=4):
                    raise Exception("Mueller matrix must be 4x4 real matrix")
            x = [x[0],x[1],x[2],x[3]]
            m = np.asarray(x,dtype=float).view(MuellerMatrix)
            return m
        if (len(x)==1):
            m = np.asarray(x,dtype=float).view(MuellerMatrix)
            m = m.reshape((4,4))
            return m
        raise Exception("Mueller matrix must be a 4x4 real matrix")

 
    def __mul__(a,b):
        """Multiply two Mueller matrices, or a Mueller matrix with 
           a Stokes vector, or multiply by a scalar 
        """
        if isinstance(b,MuellerMatrix):
            return a.dot(b)
        if isinstance(b,StokesVector):
            return StokesVector(a.dot(b))
        if isinstance(b,(float,int)):
            return MuellerMatrix(a.dot(b))
        raise Exception("Mueller matrices should only be multiplied "
                "by Mueller matrices, Stokes vectors, floats, or ints")

    __matmul__ = __mul__
    
    def __truediv__(a,b):
        "Divide a Mueller matrix by a scalar"
        if isinstance(b,int) or isinstance(b,float):
            return super().__truediv__(b)
        raise Exception("Can only divide a Mueller matrix by a float or int")

    def __pow__(a,b):
        """
        Mueller matrix to the power b.

        Parameters
        ----------

        b : float
            The power

        Returns
        -------

            : MuellerMatrix
              Matrix power operation

        """
        Q, W = np.linalg.eig(a)
        Q = Q**b
        return W.dot(np.diag(Q)).dot(np.linalg.inv(W)).real
        
    def Tmax(self):
        """
        Return the maximum transmittance.
        
        Returns
        -------

        Tmax : float
               The maximum transmittance of the MuellerMatrix
        """
        return self[0,0]+math.sqrt(self[0,1]**2+self[0,2]**2+self[0,3]**2)

    def Tmin(self):
        """
        Return the minimum transmittance.
        
        Returns
        -------

        Tmax : float
               The minimum transmittance of the MuellerMatrix
        """
        return self[0,0]-math.sqrt(self[0,1]**2+self[0,2]**2+self[0,3]**2)

    def diattenuation(self):
        """
        Return the diattenuation.
        
        Returns
        -------

        diattenuation : float
                        The diattenuation
        """
        return math.sqrt(self[0,1]**2+self[0,2]**2+self[0,3]**2)/self[0,0]

    def linear_diattenuation(self):
        """
        Return the linear diattenuation.
        
        Returns
        -------

        diattenuation : float
                        The linear diattenuation
        """
        return math.sqrt(self[0,1]**2+self[0,2]**2)/self[0,0]

    def polarization_dependent_loss(self):
        """
        Return the polarization dependent loss.

        Returns
        -------

        pdl : float
              The polarization dependent loss
        """
        return 10.*math.log(self.Tmax()/self.Tmin())/math.log(10.);

    def polarizance(self):
        """
        Return the polarizance.

        Returns
        -------

        polarizance : float
                      The polarizance
        """
        return math.sqrt(self[1,0]**2+self[2,0]**2+self[3,0]**2)/self[0,0]

    def depolarization_index(self):
        """
        Return the depolarization index.

        Returns
        -------

        PI : float
             The depolarization index
        """
        mmt = (self.T).dot(self)
        return math.sqrt((mmt[0,0]+mmt[1,1]+mmt[2,2]+mmt[3,3]-
                          self[0,0]**2)/(3*self[0,0]**2))

    def extinction_ratio(self):
        """
        Return the extinction ratio.

        Returns
        -------
        
        ER : float
             The extinction ratio
        """
        return self.Tmax()/self.Tmin()

    def rotate(self,angle):
        """
        Return a Mueller matrix rotated by an angle.

        The angle is measured clockwise looking into the beam.

        Parameters
        ----------

        angle : float
                The angle in radians

        Returns
        -------

        M : MuellerMatrix
            Mueller matrix rotated by angle
        """
        sina = math.sin(angle)
        cosa = math.cos(angle)
        R = JonesMueller([[cosa,sina],[-sina,cosa]])
        return R @ self @ R.T

    def parity(self):
        """
        Mueller matrix with a parity conversion.

        Returns
        -------

        M : MuellerMatrix
            Mueller matrix for a parity conversion
        """
        mm = self.copy()
        mm[0,2] = -mm[0,2]
        mm[0,3] = -mm[0,3]
        mm[1,2] = -mm[1,2]
        mm[1,3] = -mm[1,3]
        mm[2,0] = -mm[2,0]
        mm[2,1] = -mm[2,1]
        mm[3,0] = -mm[3,0]
        mm[3,1] = -mm[3,1]
        return mm

    def transpose(self):
        """
        Mueller matrix transpose.

        Returns
        -------

        M : MuellerMatrix
            The transposed Mueller matrix
        """
        return self.T

    def inverse(self):
        """
        Mueller matrix inverse.

        Returns
        -------

        M : MuellerMatrix
            The inverted Mueller matrix
        """
        return np.linalg.inv(self)

    def valid(self):
        """
        Test if a matrix is a Stokes-Stokes mapping matrix.

        Uses the condition described in C.R. Givens and A.B. Kostinski, 
        'A simple necessary and sufficient condition on physically realizable 
        Mueller matrices,' Journal of Modern Optics 40, 471-481 (1993).

        Returns 
        -------
            bool
            True if the matrix will always map a valid Stokes vector to a 
            valid Stokes vector; otherwise, False.
        """
        M = self + 1E-10*self[0,0]*MuellerDiagonal([1,0,0,0])
        G = MinkowskiG
        MM = G @ M.T @ G @ M
        Q, W = np.linalg.eig(MM)
        W = np.array(W.T)
        
        imax=0
        max = -1E308
        ww = W[0]
        a = (Q @ np.conj(Q)) *1E-14
        for q,w in zip(Q,W):
            if abs(q.imag)> a:
                return False
            if (abs(q)>max):
                max = abs(q)
                ww = w.copy()

        ww = ww/ww[0]
        for i in range(1,4):
            if abs(ww[i].imag) > a:
                return False
        if 1 < ww[1].real**2+ww[2].real**2+ww[3].real**2:
            return False
        return True

    def physically_valid(self):
        """ 
        Test if a matrix is physically valid.

        The function physically_valid() calculates the Mueller matrix coherency
        matrix, which is the coherency matrix for the four Jones-Mueller 
        matrices generated by the Pauli matrices. This coherency matrix must be
        a valid coherency matrix (positive semi-definite) if the matrix is the
        convex sum of Jones-Mueller matrices.  This test is a more stringent 
        test than valid() for the validity of the Mueller matrix.

        See B.N. Simon, et al, "A complete characterization of pre-Mueller and 
        Mueller matrices in polarization optics," 
        J. Opt. Soc. Am. A 27(2), 188-199 (2010).

        Returns 
        -------
            bool
            True if the matrix is physically valid, that is, if it is a convex
            sum of Jones-Mueller matrices.
        """
        h = MuellerToHermitian(self)
        Q, W = np.linalg.eig(h)
        a = (Q @ np.conj(Q)) *1E-8
        for q in Q:
            if q.real < -a:
                return False
            if abs(q.imag) > a:
                return False
        return True
    
    def Closest_NonDepolarizing(self):
        """
        Return the largest matrix in the Cloude decomposition, which may be
        interpreted as the closest non-depolarizing Mueller matrix.

        Returns
        -------

        M : MuellerMatrix
            The largest matrix in the Cloude decomposition.
        """
        M = self.Cloude_Decomposition()
        return M[0-rank]

    def normalized(self):
        """
        The normalized Mueller matrix.

        Returns
        -------

        M : MuellerMatrix
            The Mueller matrix normalized so that its [0,0] element is 1.
        """
        return self/self[0,0]

    def entropy(self):
        """
        The polarization entropy.

        Returns
        -------

        entropy : float
                  Returns the polarization entropy defined in S.R. Cloude and 
                  E. Pottier, "Concept of polarization entropy in optical
                  scattering," Opt. Eng. 34(6) 1599-1610 (1995).
        """
        h = MuellerToHermitian(self)
        Q, W =  np.linalg.eig(h)
        result = 0.
        sumQ = np.sum(Q).real
        for q in Q:
            Pi = abs(q)/sumQ
            if Pi>0:
                result += Pi*np.log(Pi)
        return -result/np.log(4.)


class StokesVector(np.ndarray):
    """
    A class for handling Stokes vectors

    Parameters
    ----------
        
    x : 4-element list or array of float or int, optional
        Default is zero Stokes vector
        
    """
    def __new__(cls,*x):
        if (len(x)==0):
            return np.asarray((0,0,0,0)).view(StokesVector)

        if (len(x)==1):
            s = np.asarray(x).view(StokesVector)
            s = s.reshape((4,))
            return s
        
        if (len(x)==4):
            s = np.asarray(x).view(StokesVector)
            return s
        
        raise Exception("Stokes vector must be 4 elements and real")

    def __mul__(a,b):
        """
        Multiply Stokes vector by a Mueller matrix, another Stokes vector 
        (inner product), or a scalar
        """
        if isinstance(b,MuellerMatrix):
            return a.dot(b)
        if isinstance(b,StokesVector):
            return float(a.dot(b))
        if isinstance(b,(float,int)):
            return StokesVector(super()*b)
    
        raise Exception("Stokes vector can only be multiplied by Mueller "
                        "matrix, a Stokes vector, or a scalar")

    __matmul__ = __mul__

    def __truediv__(a,b):
        "Divide Stokes vector by a scalar"
        if isinstance(b,(int,float)):
            return super().__truediv__(b)
        raise Exception("Stokes vector can only be divided by int or float")

    def I(self):
        return self[0]
    def Q(self):
        return self[1]
    def U(self):
        return self[2]
    def V(self):
        return self[3]
    def q(self):
        return self[1]/self[0]
    def u(self):
        return self[2]/self[0]
    def v(self):
        return self[3]/self[0]
    
    def rotate(self,angle):
        """
        Return a Stokes vector rotated by an angle.

        Parameters
        ----------

        angle : float
                Angle to rotate the Stokes vector in radians

        Returns
        -------
        rot : StokesVector
              Stokes vector rotated by angle
        """
        return JonesMueller([[math.cos(angle),math.sin(angle)],
                             [-math.sin(angle),math.cos(angle)]]) @ self

    def eta(self):
        """
        Return the principal angle of the polarization ellipse

        Returns
        -------

        eta : float
              The principal angle of the polarization ellipse
        """
        return math.atan2(self[2],self[1])/2


    def intensity(self):
        """
        Return the intensity.

        Returns
        -------
        
        I : float
            The intensity
        """
        return self[0]

    def DOLP(self):
        """
        Return the degree of linear polarization.

        Returns
        -------

        DOLP : float
               The degree of linear polarization
        """
        return np.sqrt(self[1]**2+self[2]**2)/self[0]

    def DOP(self):
        """
        Return the degree of polarization.

        Returns
        -------

        DOP : float
              The degree of polarization
        """
        return np.sqrt(self[1]**2+self[2]**2+self[3]**2)/self[0]
    
    def DOCP(self):
        """
        Return the degree of circular polarization.

        Returns
        -------

        DOCP : float
               The degree of circular polarization
        """
        return self[3]/self[0]

    def ellipticity(self):
        """
        Return the ellipticity of the field (ratio of minor to major axes).

        Returns
        -------
        
        e : float
            The ellipticity of the field (ratio of the minor to major axes 
            of the polarization ellipse.
        """
        return self[3]/(self[0]+np.sqrt(self[1]**2+self[3]**2))

    def delta(self):
        """
        Return the phase difference between the two components.

        Returns
        -------
        
        delta : float
                The phase difference between the two components (s and p) 
                in radians. Uses the polarized part of the Stokes vector.
        """
        j = self.JonesVector()
        return cmath.phase(j[1])-cmath.phase(j[0])

    def psi(self):
        """
        Return arc tangent of the amplitude ratio between the two components.

        Returns
        -------
        
        psi : float
              The arc tangent in radians of the ratio between the two 
              components of the polarized part of the light. The returned 
              value is zero for p-polarized light.  
        """
        j = self.JonesVector()
        return math.atan(np.abs(j[1])/np.abs(j[0]));

    def eccentricity(self):
        """
        Return the eccentricity.

        Returns
        -------
       
        eccentricity : float
                       The eccentricity
        """
        return np.sqrt(1-self.e()**2)

    def valid(self):
        """
        Checks if the Stokes vector is valid.

        Returns
        -------

        valid : bool
                True if the Stokes vector is valid.
        """
        return self[0]>=math.sqrt(self[1]**2+self[2]**2+self[3]**2)

    def pol_part(self):
        """
        Return the polarized part of the Stokes vector.

        Returns
        -------
        
        s : StokesVector
            The polarized part of the Stokes vector.
        """
        s = self.copy()
        s[0] = self[0]*self.DOP()
        return s;

    def unpol_part(self):
        """
        Return the unpolarized part of the Stokes vector.

        Returns
        -------
        
        s : StokesVector
            The unpolarized part of the Stokes vector.
        """
        return StokesVector(self[0]*(1-self.DOP()),0,0,0)
        return s;
    
    def JonesVector(self):
        """
        Return a Jones vector associated with the polarized part of the 
        Stokes vector.

        Returns
        -------

        j : 2 element np.array
            A Jones vector that, if converted to a Stokes vector, yields
            the polarized part of the Stokes vector.
        """
        xx = self.pol_part()
        arg = 0 if xx[3]==0. and xx[2]==0. else math.atan2(xx[3],xx[2])
        mag = np.sqrt((xx[0]-xx[1])/2.)
        return np.array([np.sqrt((xx[0]+xx[1])/2.),
                         mag*np.cos(arg) + 1j*mag*np.sin(arg)])

    def normalized(self):
        """
        Return normalized Stokes vector.

        Returns
        -------

        s : StokesVector
            The Stokes vector normalized so that its 0 element is 1.
        """
        return self/self[0]

    def __str__(self):
        return ("StokesVector(" +
                str(self[0]) + "," +
                str(self[1]) + "," +
                str(self[2]) + "," +
                str(self[3]) + ")" )

    def __repr__(self):
        return ("StokesVector(" +
                str(self[0]) + "," +
                str(self[1]) + "," +
                str(self[2]) + "," +
                str(self[3]) + ")" )
    
def JonesMueller(jones): 
    """Convert a Jones matrix to a Mueller matrix
    
    Parameters
    ----------

    jones : 2x2 list or array of int, float, or complex
            A Jones matrix

    Returns
    -------
   
    mueller : MuellerMatrix
              The Mueller matrix associated with jones.

    """
    A = np.array(((1,0,0,1),(1,0,0,-1),(0,1,1,0),(0,1j,-1j,0)))
    Ainv = np.array(((0.5,0.5,0,0),
                     (0,0,0.5,-0.5j),
                     (0,0,0.5,0.5j),
                     (0.5,-0.5,0,0)))
    return MuellerMatrix((A @ np.kron(jones,np.conj(jones)) @ Ainv).real)

def JonesStokes(jones):
    """Convert a Jones vector to a Stokes vector
    
    Parameters
    ----------

    jones : 2 element list or array of int, float, or complex

    """
    x = np.array(jones)
    if x.shape==(2,):
        cx = x.conjugate()
        return StokesVector([(x[0]*cx[0]).real+(x[1]*cx[1]).real,
                             (x[0]*cx[0]).real-(x[1]*cx[1]).real,
                             2*(cx[0]*x[1]).real,
                             2*(cx[0]*x[1]).imag])
    raise Exception("Jones vector must be a 2-element vector")


    
def MuellerToHermitian(m):
    """
    Convert a Mueller matrix into its equivalent Hermitian coherency matrix.
    
    Parameters
    ----------
    m : MuellerMatrix
        The Mueller matrix to be converted.

    Returns
    -------

    h : 4x4 complext np.array
        Hermitian coherence matrix 

    """
    h = np.ndarray((4,4),complex)

    h[0,0]=(m[0,0]+m[1,1]+m[0,1]+m[1,0])/2
    h[0,1]=complex(m[0,2]+m[1,2],m[0,3]+m[1,3])/2
    h[0,2]=complex(m[2,0]+m[2,1],-m[3,0]-m[3,1])/2
    h[0,3]=complex(m[2,2]+m[3,3],m[2,3]-m[3,2])/2
    h[1,1]=(m[0,0]-m[1,1]-m[0,1]+m[1,0])/2
    h[1,2]=complex(m[2,2]-m[3,3],-m[2,3]-m[3,2])/2
    h[1,3]=complex(m[2,0]-m[2,1],-m[3,0]+m[3,1])/2
    h[2,2]=(m[0,0]-m[1,1]+m[0,1]-m[1,0])/2
    h[2,3]=complex(m[0,2]-m[1,2],m[0,3]-m[1,3])/2
    h[3,3]=(m[0,0]+m[1,1]-m[0,1]-m[1,0])/2
    h[1,0]=np.conj(h[0,1])
    h[2,0]=np.conj(h[0,2])
    h[3,0]=np.conj(h[0,3])
    h[2,1]=np.conj(h[1,2])
    h[3,1]=np.conj(h[1,3])
    h[3,2]=np.conj(h[2,3])

    return h

def HermitianToMueller(h):
    """
    Convert a Hermitian coherency matrix to its equivalent Mueller matrix.

    Parameters
    ----------
    h : ndarray((4,4),complex)
        A Hermitian coherency matrix

    Returns
    -------
    m : MuellerMatrix
        The Mueller matrix form of the coherency matrix.

    """
    mm = MuellerMatrix()
    mm[0,0] = (h[0,0]+h[1,1]+h[2,2]+h[3,3]).real/2
    mm[0,1] = (h[0,0]-h[1,1]+h[2,2]-h[3,3]).real/2
    mm[0,2] = (h[0,1]+h[1,0]+h[2,3]+h[3,2]).real/2
    mm[0,3] = (h[0,1]-h[1,0]+h[2,3]-h[3,2]).imag/2
    mm[1,0] = (h[0,0]+h[1,1]-h[2,2]-h[3,3]).real/2
    mm[1,1] = (h[0,0]-h[1,1]-h[2,2]+h[3,3]).real/2
    mm[1,2] = (h[0,1]+h[1,0]-h[2,3]-h[3,2]).real/2
    mm[1,3] = (h[0,1]-h[1,0]-h[2,3]+h[3,2]).imag/2
    mm[2,0] = (h[0,2]+h[2,0]+h[1,3]+h[3,1]).real/2
    mm[2,1] = (h[0,2]+h[2,0]-h[1,3]-h[3,1]).real/2
    mm[2,2] = (h[0,3]+h[3,0]+h[1,2]+h[2,1]).real/2
    mm[2,3] = (h[0,3]-h[3,0]-h[1,2]+h[2,1]).imag/2
    mm[3,0] = (h[2,0]-h[0,2]-h[1,3]+h[3,1]).imag/2
    mm[3,1] = (h[2,0]-h[0,2]+h[1,3]-h[3,1]).imag/2
    mm[3,2] = (h[3,0]-h[0,3]+h[2,1]-h[1,2]).imag/2
    mm[3,3] = (h[0,3]+h[3,0]-h[1,2]-h[2,1]).real/2

    return mm


def Cloude_Decomposition(M):
    """
    Return a list of four non-depolarizing Mueller matrices that sum to the 
    Mueller matrix.

    Parameters
    ----------
        M : MuellerMatrix
            Mueller matrix to be decomposed
    
    Returns
    -------
        M1, M2, M3, M4 : MuellerMatrix 
                         The decomposed Jones-Mueller matrices sorted so that
                         M1[0,0] >= M2[0,0] >= M3[0,0] >= M4[0,0]

    """
    h = MuellerToHermitian(M)
    Q, W = np.linalg.eig(h)
    Winv = np.linalg.inv(W)
    
    MM = [HermitianToMueller(Q[i] * np.outer( W[:,i] , Winv[i,:] ))
         for i in range(4)]
    
    for i in range(0,3):
        for j in range(3,i,-1):
            if MM[j-1][0,0]<MM[j][0,0]:
                MM[j-1], MM[j] = MM[j], MM[j-1]
    return MM

MuellerMatrix.Cloude_Decomposition = Cloude_Decomposition

LeviCivita = np.array([[[ 0,  0,  0],
                        [ 0,  0,  1],
                        [ 0, -1,  0]],
                       [[ 0,  0, -1],
                        [ 0,  0,  0],
                        [ 1,  0,  0]],
                       [[ 0,  1,  0],
                        [-1,  0,  0],
                        [ 0,  0,  0]]])

def Lu_Chipman_Decomposition(M):
    dep, ret, diatten = SCATPY.Lu_Chipman_Decomposition(*(M.flatten().tolist()))
    return MuellerMatrix(dep), MuellerMatrix(ret), MuellerMatrix(diatten)
    
def Test_Lu_Chipman_Decomposition(M):
    """
    Return a depolarizer, a retarder, and a diattenuator whose ordered product is the
    Mueller matrix. 

    See: S.-Y. Lu and R.A. Chipman, "Interpretation of Mueller matrices based on 
    polar decomposition," J. Opt. Soc. Am. A **13**, 1106-1113 (1996).

    Parameters
    ----------
       M : MuellerMatrix
           The matrix to be decomposed

    Returns
    -------
       depolarizer : MuellerMatrix
       retarder : MuellerMatrix
       diattenuator : MuellerMatrix
            M = depolarizer @ retarder @ diattenuator
    """
    diattenuator = MuellerMatrix()
    retarder = MuellerMatrix()
    depolarizer = MuellerMatrix()
    
    # The diattenuator takes the net transmittance of the matrix.
    # See L&C, Eq. (18)
    Tu = M[0,0]

    # If the transmittance is zero, then the decomposition is trivial.
    if (Tu==0):
        return (MuellerMatrix(np.identity(4)),
                MuellerMatrix(np.identity(4)),
                MuellerMatrix())

    # From L&C, Eq. (36), Darrow and Parrow are the diattenuation and
    # polarizance, respectively.
    Darrow = np.array( M[0,1:]/Tu )
    Parrow = np.array( M[1:,0]/Tu )
    D = (np.sum(Darrow**2))**0.5

    # Thus, the diattenuator is given by L&C, Eq. (18),
    if D>1E-15:
        diattenuator = MuellerMatrix()
        diattenuator[0,0] = 1
        diattenuator[1:,0] = Darrow
        diattenuator[0,1:] = Darrow
        dd = math.sqrt(1-D**2)
        
        diattenuator[1:,1:] = dd*np.identity(3) + ((1-dd)/D**2)*np.outer(Darrow,Darrow)

        del dd                    
    else:
        diattenuator = MuellerUnit()

    diattenuator = diattenuator*Tu

    norm = np.linalg.norm
    cross = np.cross
    sqrt = np.sqrt
                                    
    inverse = np.linalg.inv

    # If the matrix is non-depolarizing, then use
    if abs(1-M.depolarization_index()) < 1E-8:
        depolarizer=MuellerUnit()
        # If the diattenuation is unity (or very close!)...
        if 1-D < 1E-8:
            # Get the retardance vector direction...
            Rarrow = cross(Parrow,Darrow)
            normRarrow = norm(Rarrow)
            Rhat = unit(Rarrow) if normRarrow!=0 else perpto(Parrow,[0,0,1])   
            cosR = Parrow*Darrow/norm(Parrow)/norm(Darrow)
            sinR = sqrt(1-cosR**2) if abs(cosR)<1 else 0
            
            # The retarder is determined from L&C, Eqs. (14) and (15)...
            retarder[0,0]=1
            retarder[0,1:4]=0
            retarder[1:4]=0
            retarder[1:4,1:4] = cosR*np.identity(3) + (1-cosR)*np.outer(Rhat,Rhat) + sinR*LeviCivita @ Rhat

        else:
            # If the matrix does not have unit diattenutation, then
            # just use L&C, Eq. (35)...
            retarder = M*np.linalg.inv(diattenuator)

        return depolarizer, retarder, diattenuator

    # L&C, Eq. (47)...
    Mprime = M.dot(np.linalg.inv(diattenuator))

    # L&C, Eq. (48)...
    mprime = Mprime[1:4,1:4]

    # Also from L&C, Eq. (48)...
    PDelta = Mprime[1:4,0]

    # Get the determinant of mprime...
    mprimedeterminant = np.linalg.det(mprime)

    # If mprime is not singular...
    if mprimedeterminant!=0: 
        # To evaluate L&C, Eq. (52), we need the eigenvalues of
        # m'(m')^T ...
        mmT = mprime.dot(mprime.T)

        # Calculate eigenvalues...
        Q, W = np.linalg.eig(mmT)

        # These are the square roots of the eigenvalues
        lambda1 = sqrt(Q[0].real)
        lambda2 = sqrt(Q[1].real)
        lambda3 = sqrt(Q[2].real)

        ident = np.diag([1.,1.,1.])

        # Finally, this is L&C, Eq. (52)...
        mDelta = (
            np.linalg.inv(mmT+(lambda1*lambda2+lambda2
                               *lambda3+lambda3*lambda1)*ident).dot(
            ((lambda1+lambda2+lambda3)*mmT+lambda1*lambda2*lambda3*ident))
            )

        # Use the right sign, using the determinant...
        if mprimedeterminant<0:
            mDelta = -mDelta

        # Thus, the depolarizer is given by L&C, Eq. (48)...
        depolarizer[0,0] = 1
        depolarizer[0,1:4] = 0
        depolarizer[1:4,0] = PDelta
        depolarizer[1:4,1:4] = mDelta

        # And the retarder is determined by L&C, Eq. (53)...
        retarder = np.linalg.inv(depolarizer)*Mprime

    else:
        # If the matrix is singular, then we have to use the
        # formalism described in L&C, Appendix B...

        # First, perform the singular value decomposition
        # of mprime...
        U, S, V = np.linalg.svd(mprime)

        # The singular values are...
        lambda1 = S[0]
        lambda2 = S[1]
        lambda3 = S[2]


        mR = np.diag([0.,0.,0.])
        # How one determines the retardance depends upon how many of the
        # singular values are zero...
        if abs(lambda2/lambda1)>1E-15:
            # One of them is zero...
            mR = np.outer(V[:,0],U[:,0])+np.outer(V[:,2],U[:,2])+np.outer(V[:,2],U[:,2])
        elif lambda1!=0:
            # Two of them are zero...
            v = V[:,0]
            u = U[:,0]

            pv = perpto(v,[1,0,0])
            ppv = perpto(pv,v)
            pu = perpto(u,[1,0,0])
            ppu = perpto(pu,u)
            # Essentially, L&C, Eq. (B9)...
            mR = np.outer(v,u)+np.outer(pv,pu)+np.outer(ppv,ppu)
        else:
            # The case of all of them zero is trivial...
            mR = np.diag([1.,1.,1.])

        # The full retarder Mueller matrix is...
        retarder[0,0] = 1
        retarder[1:4,0] = 0
        retarder[0,1:4] = 0
        retarder[1:4,1:4] = mR

        # The depolarizer is determined from...
        # (The matrix mR is always invertible.)
        mDelta = mprime.dot(np.linalg.inv(mR))

        depolarizer[0,0] = 1
        depolarizer[0,1:4] = 0
        depolarizer[1:4,0] = PDelta
        depolarizer[1:4,1:4] = mDelta

    return depolarizer, retarder, diattenuator

MuellerMatrix.Lu_Chipman_Decomposition = Lu_Chipman_Decomposition

def Reverse_Lu_Chipman_Decomposition(M):
    """
    Return a diattenuator, retarder, and depolarizer whose ordered product is the
    Mueller matrix. 

    See: S.-Y. Lu and R.A. Chipman, "Interpretation of Mueller matrices based on 
    polar decomposition," J. Opt. Soc. Am. A **13**, 1106-1113 (1996)
    and J. Morio and G. Goudail, "Influence of the order of diattenuator, retarder, 
    and polarizer in polar decomposition of Mueller matrices," 
    Opt. Lett. **29**, 2234-2236 (2004).

    Parameters
    ----------
       M : MuellerMatrix
           The matrix to be decomposed

    Returns
    -------
       diattenuator : MuellerMatrix
       retarder : MuellerMatrix
       depolarizer : MuellerMatrix
            M = diattenuator @ retarder @ depolarizer
    """
    try:
        diattenuator = MuellerMatrix()
        retarder = MuellerMatrix()
        depolarizer = MuellerMatrix()

        Darrow = np.linalg.inv(M[1:,1:]).T @ M[0,1:]
        D2 = Darrow @ Darrow
        D = np.sqrt(D2)
    
        diattenuator0 = MuellerMatrix()
        diattenuator0[0,0] = 1
        diattenuator0[1:,0] = Darrow
        diattenuator0[0,1:] = Darrow
        dd = np.sqrt(1 - D2)

        diattenuator0[1:,1:] = dd * np.identity(3) + ((1 - dd) / D2) * np.outer(Darrow, Darrow)

        Mprime = np.linalg.inv(diattenuator0) @ M
        Tu = Mprime[0,0]
        diattenuator = diattenuator0 * Tu

        U, S, VT = np.linalg.svd(np.array(Mprime[1:,1:]) / Tu)  

        mR = U @ VT
        mDelta = VT.T @ np.diag(S) @ VT
        if np.linalg.det(mR) < 0:
            mR = -mR
            mDelta = -mDelta

        Parrow = np.linalg.inv(mR) @ Mprime[1:,0] / Tu

        retarder[0,0] = 1
        retarder[0,1:] = 0
        retarder[1:,0] = 0
        retarder[1:,1:] = mR

        depolarizer[0,0] = 1
        depolarizer[0,1:] = 0
        depolarizer[1:,0] = Parrow
        depolarizer[1:,1:] = mDelta

        return diattenuator, retarder, depolarizer
    except:
        pass
    raise Exception("Reverse Lu-Chipman Decomposition not possible")

MuellerMatrix.Reverse_Lu_Chipman_Decomposition = (
                                    Reverse_Lu_Chipman_Decomposition)

class CharacterizedMueller():
    """
    Characterize the Mueller matrix according to 
    
    N. Ghosh, M.F.G. Wood, and I.A. Vitkin,
    'Mueller matrix decomposition for extraction of individual polarization 
    parameters from complex turbid media exhibiting multiple scattering, optical
    activity, and linear birefringence,'' J. Biomedical Opt. 13, 044036 (2008).
    
    A few other parameters are also calculated.
    
    Parameters
    ----------
    M : MuellerMatrix
        The Mueller matrix to be characterized
    
    Attributes
    ----------
    Mdepol : MuellerMatrix
             Depolarizer in Lu-Chipman decomposition
    
    Mret : MuellerMatrix
           Retarder in Lu-Chipman decomposition
    
    Mdiatten : MuellerMatrix
               Diattenuator in Lu-Chipman decomposition
    
    DiattenuationVector : numpy.array
                          The 3-element diattenuation vector
    
    Diattenuation : float
                    Diattenuation
    
    CircularDiattenuation : float
                            Circular diattenuation
    
    LinearDiattenuation : float
                          Linear diattenuation
    
    DiattenuationAngle : float
                         Angle of diattenuation in radians
    
    Polarizance : numpy.array
                  The 3-element polarizance vector
    
    DepolarizationCoefficient : float
                                Depolarization coefficient 
                                (1 = nondepolarizing, 0 = depolarizing)
    
    LinearRetardance : float
                       Linear retardance in radians
                       
    OpticalRotation : float
                      Optical rotation in radians
                      
    Retardance : float
                 Total retardance in radians
    
    RetardanceAngle : float
                      Angle of retardance in radians
                          
    """
    def __init__(self,M):
        Mdepol,Mret,Mdiatten = Lu_Chipman_Decomposition(M)
        
        self.Mdepol = Mdepol
        self.Mret = Mret
        self.Mdiatten = Mdiatten

        dv = np.array(Mdiatten[0,1:]/Mdiatten[0,0])
        self.DiattenuationVector = dv
        self.Diattenuation = math.sqrt(sum(dv**2))
        self.CircularDiattenuation = dv[2] 
        self.LinearDiattenuation = math.sqrt(sum(dv[:2]**2))
        self.DiattenuationAngle = math.atan2(Mdiatten[2,0],Mdiatten[1,0])/2
        
        self.PolarizanceVector = np.array(Mdepol[1:4,0])
        self.Polarizance = math.sqrt(sum(self.PolarizanceVector**2))
        self.DepolarizationCoefficient = abs(np.trace(Mdepol)-1)/3

        self.LinearRetardance = math.acos(math.sqrt((Mret[1,1]+Mret[2,2])**2+
                                                    (Mret[2,1]-Mret[1,2])**2)-1)
        self.OpticalRotation = math.atan((Mret[2,1]-Mret[1,2])/
                                         (Mret[1,1]+Mret[2,2]))/2
        self.Retardance = math.acos(np.trace(Mret)/2-1)
        psi = self.OpticalRotation
        Mcircret = MuellerMatrix([[1,0,0,0],
                                  [0,math.cos(2*psi),math.sin(2*psi),0],
                                  [0,-math.sin(2*psi),math.cos(2*psi),0],
                                  [0,0,0,1]])
        Mlinret = Mret @ Mcircret.inverse()
        self.RetardanceAngle = math.atan2((Mlinret[3,1]-Mlinret[1,3]),
                                     (Mlinret[2,3]-Mlinret[3,2]))/2

    def __str__(self):
        s = ("Mdepol = \n" + str(self.Mdepol)
             + "\nMret = \n" + str(self.Mret)
             + "\nMdiatten = \n" + str(self.Mdiatten)
             + "\nDiattenuationVector = " + str(self.DiattenuationVector)
             + "\nDiattenuation = " + str(self.Diattenuation) 
             + "\nCircularDiattenuation = " + str(self.CircularDiattenuation)
             + "\nLinearDiattenuation = " + str(self.LinearDiattenuation)
             + "\nDiattenuationAngle = " + str(self.DiattenuationAngle)
             + " rad (" + str(self.DiattenuationAngle/deg) + " deg)"
             + "\nPolarizanceVector = " + str(self.PolarizanceVector)
             + "\nPolarizance = " + str(self.Polarizance)
             + "\nDepolarizationCoefficient = "
             + str(self.DepolarizationCoefficient)
             + "\nLinearRetardance = " + str(self.LinearRetardance)
             + " rad (" + str(self.LinearRetardance/deg) + " deg)"
             + "\nOpticalRotation = " + str(self.OpticalRotation)
             + " rad (" + str(self.OpticalRotation/deg) + " deg)"
             + "\nRetardance = " + str(self.Retardance)
             + " rad (" + str(self.Retardance/deg) + "deg)"
             + "\nRetardanceAngle = " + str(self.RetardanceAngle)
             + " rad (" + str(self.RetardanceAngle/deg) + " deg)")
        return s
    

def Symmetric_Decomposition(M):
    """
    Perform the symmetric decomposition described by Ossikovski.

    See: R. Ossikovski, J. Opt. Soc. Am. A **26**, 1109-1118 (2009).

    Parameters
    ----------
       M : MuellerMatrix

    Returns
    -------
       diatten2 : MuellerMatrix
       ret2 : MuellerMatrix
       depol : MuellerMatrix
       ret1 : MuellerMatrix
       diatten1 : MuellerMatrix
             M = diatten2 @ ret2 @ depol @ ret1 @ diatten1
    """
    def diattenFromVec(D):
        """
        Construct the diattenuator from the diattenuation veotor.
        """
        MD = np.zeros((4,4))
        MD[0,0] = 1
        for i in range(1,4):
            MD[0,i] = D[i]
            MD[i,0] = D[i]
        Dhat = np.array([D[1],D[2],D[3]])
        Dsqr = Dhat @ Dhat
        Dhat = Dhat/np.sqrt(Dsqr)
        if Dsqr<1:
            mD = (np.sqrt(1-Dsqr)*np.diag([1,1,1])+
                  (1-np.sqrt(1-Dsqr))*np.outer(Dhat,Dhat))
        else:
            mD = np.outer(Dhat,Dhat)

        for i in range(3):
            for j in range(3):
                MD[i+1,j+1] = mD[i,j]
        return MD

    def retFromSubMatrix(m):
        """
        Constructs the retarder from the 3x3 submatrix
        """
        MR = np.zeros((4,4))
        MR[0,0]=1
        for i in range(3):
            for j in range(3):
                MR[i+1,j+1] = m[i,j]
        return MR

    def FindValid(Q,W):
        """
        Given the eigenvalues Q and eigenvector matrix W, returns
        the eigenvector associated with the valid Stokes vector
        """
        W = W.T

        imax=0
        max = -1E308
        ww = W[0]
        a = (Q @ np.conj(Q)) *1E-14
        for q,w in zip(Q,W):
            if abs(q.imag)> a:
                raise ValueError("Not a valid Mueller matrix")
            if (abs(q)>max):
                max = abs(q)
                ww = w.copy()

        if ww[0].real**2 < ww[1].real**2+ww[2].real**2+ww[3].real**2:
            raise ValueError("Not a valid Mueller matrix")

        return ww

    # Step 1 
    M = np.array(M) + np.diag([1,0,0,0])*M[0,0]*1E-10
    G = np.array(MinkowskiG)
    MTGMG = M.T @ G @ M @ G
    Q, W = np.linalg.eig(MTGMG)

    S1 = FindValid(Q,W)
    S1 = S1/S1[0]
    S2 = M @ G @ S1
    S2 = S2/S2[0]

    # Step 2
    diatten1 = diattenFromVec(S1)
    diatten2 = diattenFromVec(S2)

    # Step 3
    Mprime = np.linalg.inv(diatten2) @ M @ np.linalg.inv(diatten1)

    # Step 4
    mprime = Mprime[1:4,1:4]
    mR2, mdepol, mR1T = np.linalg.svd(mprime)

    ret1 = retFromSubMatrix(mR1T)
    ret2 = retFromSubMatrix(mR2)

    depol = np.diag([Mprime[0,0],mdepol[0],mdepol[1],mdepol[2]])

    return (MuellerMatrix(diatten2),
            MuellerMatrix(ret2),
            MuellerMatrix(depol),
            MuellerMatrix(ret1),
            MuellerMatrix(diatten1))

MuellerMatrix.Symmetric_Decomposition = Symmetric_Decomposition

def MuellerLog(M):
    """
    Return matrix logarithm (logarithmic decomposition).

    Returns
    -------
    4x4 numpy.array 
        The matrix logarithm of M 
    """
    Q, W = np.linalg.eig(np.array(M))
    QQ = []
    for q in Q:
        if q<=0:
            QQ.append(np.log(q*(1+1E-16j)))
        else:
            QQ.append(np.log(q))  
    L = W @ np.diag(QQ) @ np.linalg.inv(W)
    isreal = True
    for el in L.flatten():
        if abs(el.imag)>1E-14:
            isreal = False
    if isreal: return L.real
    else: return L

MuellerMatrix.MuellerLog = MuellerLog

def MuellerExp(L):
    """
    Return matrix exponential.

    Returns
    -------

    L : 4x4 np.array 
        The matrix exponent of M
    """
    Q, W = np.linalg.eig(np.array(L))
    Q = np.exp(Q)
    M = W @ np.diag(Q) @ np.linalg.inv(W)
    isreal = True
    for m in M.flatten():
        if abs(m.imag)>1E-14:
            isreal = False
    if isreal: return M.real
    else: return M

MuellerMatrix.MuellerExp = MuellerExp

MinkowskiG = MuellerMatrix(np.diag([1,-1,-1,-1]))

PauliMatrices = [np.array([[1,0],
                           [0,1]]),
                 np.array([[1,0],
                           [0,-1]]),
                 np.array([[0,1],
                           [1,0]]),
                 np.array([[0,-1j],
                           [1j,0]])]

def JonesZero():
    """
    Return a zero Jones matrix
    """
    return np.array([[0,0],[0,0]])

def JonesUnit():
    """
    Return a unit Jones matrix
    """
    return np.array([[1,0],[0,1]])

def JonesRotator(angle):
    """
    Return a Jones matrix that rotates the polarization by angle"

    Parameters
    ----------

    angle : float
            Angle that matrix will rotate in radians

    Returns
    -------

    R : 2x2 np.array
        The rotation matrix 
    """
    cosa = np.cos(angle)
    sina = np.sin(angle)
    return np.array([[cosa,sina],[-sina,cosa]])

def MuellerRotator(angle):
    """
    Return a Mueller matrix that rotates the polarization by angle"

    Parameters
    ----------

    angle : float
            Angle that matrix will rotate in radians

    Returns
    -------

    R : MuellerMatrix
        The rotation matrix 
    """
    cos2a = np.cos(2*angle)
    sin2a = np.sin(2*angle)
    return MuellerMatrix(((1,0,0,0),
                          (0,cos2a,sin2a,0),
                          (0,-sin2a,cos2a,0),
                          (0,0,0,1)))

def JonesBasis(angle=0, DOCP=None, jones=None):
    """
    Return the basis vectors for a Jones Matrix using a Jones vector or a 
    degree of circular polarization and an angle. The returned vectors are 
    orthogonal to one another and of unit intensity. This function
    is primarily used by JonesRetarder and JonesDiattenuator
    
    Parameters
    ----------
    
    angle : float
            Angle (in radians) of axis (default is 0)
    
    DOCP : float
           Degree of circular polarization of the axis.
           Cannot define both DOCP and jones. 
    
    jones : 2-element list of complex
            Jones vector associated with axis
            Cannot define both DOCP and jones. (Default is [1,0])
    
    Returns
    -------
    
    jones1, jones2 : 2-element lists of complex
                     Two Jones vectors that are mutually orthogonal 
                     and of unit intensity
    """
    if jones == None:
        try:
            jones1 = np.array([1,(1/DOCP-np.sqrt(1/DOCP**2-1))*1j])
        except:
            jones1 = np.array([1,0])
    else:
        if DOCP == None:
            jones1 = jones/np.sqrt(jones @ np.conj(jones))
        else:
            raise Error("DOCP and jones both defined")
    
    jones2 = np.array([np.conj(jones1[1]),-np.conj(jones1[0])])
    jones1 = jones1/np.sqrt(jones1 @ np.conj(jones1))
    jones2 = jones2/np.sqrt(jones2 @ np.conj(jones2))

    R = JonesRotator(angle)
    return (jones1 @ R,jones2 @ R)

def JonesRetarder(phase=None, angle=0, DOCP=None, jones=None):
    """
    Return the Jones matrix for a homogeneous retarder with a given phase 
    and axes. 
    
    Parameters
    ----------
    
    angle : float
            Angle (in radians) of axis (default is 0)
    
    DOCP : float
           Degree of circular polarization of the axis.
           Cannot define both DOCP and jones. 
    
    jones : 2-element list of complex
            Jones vector associated with axis
            Cannot define both DOCP and jones. (Default is [1,0])
    
    Returns
    -------
    
    jones : 2x2 numpy array of complex
            Jones matrix retarder
    
    """
    j1, j2 = JonesBasis(angle, DOCP, jones)
    return np.outer(j1,np.conj(j1)) + np.outer(j2,np.conj(j2))*np.exp(1j*phase)

        
def JonesDiattenuator(diatten=None, angle=0, DOCP=None, jones=None):
    """
    Return the Jones matrix for a homogeneous diattenuator with a given 
    phase and axes. 
    
    Parameters
    ----------
    
    angle : float
            Angle (in radians) of axis (default is 0)
    
    DOCP : float
           Degree of circular polarization of the axis.
           Cannot define both DOCP and jones. 
    
    jones : 2-element list of complex
            Jones vector associated with axis
            Cannot define both DOCP and jones. (Default is [1,0])
    
    Returns
    -------
    
    jones : 2x2 numpy array of complex
            Jones matrix diattenuator
    
    """
    j1, j2 = JonesBasis(angle, DOCP, jones)
    return np.outer(j1,np.conj(j1)) + np.outer(j2,np.conj(j2))*np.sqrt(diatten)

def MuellerRetarder(phase=0, angle=0, DOCP=None, jones=None):
    """
    Return the Mueller matrix for a homogeneous retarder with a given 
    phase and axis. 
    
    Parameters
    ----------
    
    angle : float
            Angle (in radians) of axis (default is 0)
    
    DOCP : float
           Degree of circular polarization of the axis.
           Cannot define both DOCP and jones. 
    
    jones : 2-element list of complex
            Jones vector associated with axis
            Cannot define both DOCP and jones. (Default is [1,0])
    
    Returns
    -------
    
    jones : 2x2 numpy array of complex
            Jones matrix retarder
    
    """
    return JonesMueller(JonesRetarder(phase,angle,DOCP,jones))

def MuellerDiattenuator(diatten=0, angle=0, DOCP=None, jones=None):
    """
    Return the Mueller matrix for a homogeneous diattenuator with a 
    given phase and axis. 
    
    Parameters
    ----------
    
    angle : float
            Angle (in radians) of axis (default is 0)
    
    DOCP : float
           Degree of circular polarization of the axis.
           Cannot define both DOCP and jones. 
    
    jones : 2-element list of complex
            Jones vector associated with axis
            Cannot define both DOCP and jones. (Default is [1,0])
    
    Returns
    -------
    
    MuellerMatrix diattenuator
    
    """    
    return JonesMueller(JonesDiattenuator(diatten,angle,DOCP,jones))

JonesPolarizer = JonesDiattenuator
MuellerPolarizer = MuellerDiattenuator

def MuellerZero():
    """
    Return a zero Mueller matrix

    Returns
    -------

    m : MuellerMatrix
        Zero matrix
    """
    return MuellerMatrix(np.zeros((4, 4)))

def MuellerUnit(attenuation=1):
    """
    Return a unit Mueller matrix

    Arguments
    --------
    attenuation : float, optional
                  Default is 1
    """
    return MuellerMatrix(attenuation * np.identity(4))

def MuellerDiagonal(a):
    """
    Return a diagonal Mueller matrix

    Arguments
    ---------
    a : 4 element list or array of float 
    """
    return MuellerMatrix(np.diag(a))

def StokesZero():
    """
    Return zero Stokes vector

    """
    return StokesVector(np.zeros(4))
    
def Polarization(*args, **kwargs):
    """
    Return a StokesVector with a given name
   
    Parameters
    ----------

    args : str or float
           Can be any of:
               `P` or `p` for p-polarization
               `S` or `s` for s-polarization
               `L` or `l` for LCP-polarization
               `R` or `r` for RCP-polarization
               `U` or `u` for unpolarized
               any float for linear polarization at angle `name` in radians
    
    state : str
            Any of the aboves strings that can be used in args

    I : float, optional
        The intensity. (Default is 1)
    
    DOP : float, optional
          The degree of polarization. (Default is 1)

    angle : float, optional
            The principal angle of the polarization. (Default is set by
            args or state.)

    ellipticity : float
                  The ellipticity of the polarization

    Returns
    -------

    s : StokesVector
        The Stokes vector with the given parameters
    """
    state = None
    sensitivity = False
    if 'sensitivity' in kwargs:
        sensitivity = kwargs.pop('sensitivity')

    try:
        if len(args)!=0:
            s = StokesVector(*args)
            if sensitivity:
                return s.pol_part()/2+s.unpol_part()
            else:
                return s
    except Exception:
        pass

    if len(args)==1:
        state = args[0]
    
    I = 1
    p = 1
    chi = 0
    
    if 'state' in kwargs:
        state = kwargs.pop('state')
    if 'I' in kwargs:
        I = kwargs.pop('I')
    if 'DOP' in kwargs:
        p = kwargs.pop('DOP')
    if 'angle' in kwargs:
        psi = kwargs.pop('angle')
    if 'ellipticity' in kwargs:
        chi = kwargs.pop('ellipticity')
    if len(kwargs) != 0:
        raise Exception("Unrecognized keyword")

    if sensitivity:
        s0 = (1-p/2)
        s1 = p/2
        p = 1
    else:
        s0 = 1
        s1 = 1
        
    if state == "P" or state == 'p':
        return StokesVector(I*s0,-I*p*s1,0,0)
    if state == "S" or state == 's':
        return StokesVector(I*s0,I*p*s1,0,0)
    if state == "U" or state == 'u':
        return StokesVector(I,0,0,0)
    if state == "L" or state == 'l':
        return StokesVector(I*s0,0,0,I*p*s1)
    if state == "R" or state == 'r':
        return StokesVector(I*s0,0,0,-I*p*s1)
    
    return StokesVector(I*s0,
                        I*p*s1*math.cos(2*psi)*math.cos(2*chi),
                        I*p*s1*math.sin(2*psi)*math.cos(2*chi),
                        I*p*s1*math.sin(2*chi))


def Sensitivity(*args, **kwargs):
    """
    Return a StokesVector with a given name that 
    corresponds to a sensitivity.  
   
    Parameters
    ----------

    name : str or float
           Can be any of:
               `P` or `p` for p-polarization
               `S` or `s` for s-polarization
               `L` or `l` for LCP-polarization
               `R` or `r` for RCP-polarization
               `U` or `u` for unpolarized
               any float for linear polarization at angle `name`
    """
    if 'sensitivity' not in kwargs:
        kwargs['sensitivity'] = True

    return Polarization(*args, **kwargs)

    



#
# Test script for pySCATMECH. This will create lots of graphs and output. Should not crash!
#

print("""
This script exercises several features in each module of pySCATMECH.
All temporary files should be deleted at the end. 
Numerous windows will open up. Close each to continue.
Press return now to start.""")
input()

def WORKING():
    print("Working...",end="",flush=True)

def DONE():
    print("Done")


print("==========================")
print("Testing pySCATMECH.mueller")
print("==========================")

from pySCATMECH.mueller import *
import random 
import matplotlib.pyplot as plt

print("Creating some unit MuellerMatrix")
m1 = MuellerMatrix([[1,0,0,0],[0,1,0,0],[0,0,1,0],[0,0,0,1]])
m2 = MuellerMatrix([1,0,0,0],[0,1,0,0],[0,0,1,0],[0,0,0,1])
m3 = MuellerMatrix(1,0,0,0,0,1,0,0,0,0,1,0,0,0,0,1)
m4 = JonesMueller([[1,0],[0,1]])
m5 = MuellerUnit()
print(m5)

print("Mueller matrix maps Stokes vectors to Stokes vectors:",m1.valid())
print("Mueller matrix is a convex sum of non-depolarizing Mueller matrices:",m1.physically_valid())

# Returns a random number with a normal distribution...
ran = lambda : random.gauss(0,1)

print("Creating 20 totally random Mueller matrices.")
m1 = [MuellerMatrix([[ran(), ran(), ran(), ran()],
                    [ran(), ran(), ran(), ran()],
                    [ran(), ran(), ran(), ran()],
                    [ran(), ran(), ran(), ran()]]) for i in range(20)]

print("Number valid() is True:", sum([m.valid() for m in m1]))
print("Number physically_valid() is True:", sum([m.physically_valid() for m in m1]))

# Returns a random complex number with a normal distribution...
ranc = lambda : random.gauss(0,1)+random.gauss(0,1)*1j

print("Creating 20 random Jones-Mueller matrices")
m1 = [JonesMueller([[ranc(), ranc()], [ranc(), ranc()]]) for i in range(20)]

# See if they are valid...
print("Number value() is True:", sum([m.valid() for m in m1]))
print("Number physically valid() is True:", sum([m.physically_valid() for m in m1]))

def RandomMuellerMatrix(n):
    """
    Creates a random Mueller matrix by averaging `n` Jones-Mueller matrices.
    """
    ranc = lambda : (random.gauss(0,1)+random.gauss(0,1)*1j)/2
    return sum([JonesMueller([[ranc(),ranc()],[ranc(),ranc()]]) for i in range(n)])/n

print("A random Mueller matrix generated from sum of four random nondepolarizing Mueller matrices:")
m = RandomMuellerMatrix(4)
print(m)
print("m is a valid Stokes to Stokes mapper:",m.valid())
print("m is the sum of Jones-Mueller matrices:",m.physically_valid())

print("Unpolarized Stokes vector:", StokesVector((1,0,0,0)) )

print("Unpolarized Stokes vector:", StokesVector(1,0,0,0) )

print("s-polarized unit polarization from a Jones vector:", JonesStokes((1.,0.)) )

print("p-polarized unit Stokes vector:", Polarization('p') )

print("s-polarized unit Stokes vector:", Polarization(state = 's') )

print("Left-handed circularly polarized unit Stokes vector:", Polarization("L") )

print("Right-handed circularly polarized Stokes vector with intensity 2:", Polarization("R", I=2) )

print("Unpolarized Stokes vector:", Polarization("U") )

print("Unpolarized Stokes vector:", Polarization('s', DOP=0) )

print("Linear polarization at 45 degrees:", Polarization(angle = 45*deg) )

print("Elliptically polarized radiation:", Polarization(angle = 34*deg, ellipticity = 20*deg) )

print("Partially polarized radiation:", Polarization("S", DOP = 0.8) )

print("Partially polarized elliptical polarized radiation with intensity 3:\n", 
      Polarization(angle = 34*deg, ellipticity = 20*deg, DOP = 0.7, I = 3) )

def RandomStokesVector(n):
    ranc = lambda : (random.gauss(0,1) + random.gauss(0,1)*1j)/2
    return sum([JonesStokes([ranc(),ranc()]) for i in range(n)])/n

print("Creating random Stokes vectors with the sum of n random polarized Stokes vectors")
plt.figure()
plt.plot(range(1,100),[RandomStokesVector(i).DOP() for i in range(1,100)])
plt.title("Degree of polarization vs. number of Jones-Stokes vectors")
plt.xlabel("n")
plt.ylabel("DOP")
plt.show()


jmax = 100
imax = 100
print("Creating",imax*jmax,"random Mueller matrices")

WORKING()

mCD00sums = []
for i in range(1,imax):
    mCD00sum = np.array([0,0,0,0])
    for j in range(jmax):
        M = RandomMuellerMatrix(i)
        mCD = Cloude_Decomposition(M)
        mCD00sum = mCD00sum + np.array([m[0,0] for m in mCD])
    mCD00sums.append(mCD00sum/jmax)
DONE()
plt.figure()
plt.plot(mCD00sums)
plt.xlabel("Number of Jones-Mueller matrices averaged")
plt.ylabel("Cloude decomposition")
plt.show()

def printMatrix(name,M):
    print("\n",name,"=\n")
    print(M)

print("Creating a random Mueller matrix.")
M = RandomMuellerMatrix(4)

MCD = M.Cloude_Decomposition()
printMatrix("M",M)

print("\nCloude decomposition")
print("--------------------\n")

for m in MCD:
    print(m,"\n")

depol, ret, diatten = Lu_Chipman_Decomposition(M)
print("\nLu-Chipman Decomposition")
print("------------------------\n")

printMatrix("Depolarizer",depol)
printMatrix("Retarder",ret)
printMatrix("Diattenuator",diatten)
printMatrix("The product Depolarizer . Retarder .  Diattenuator",depol @ ret @ diatten)

diatten, ret, depol = Reverse_Lu_Chipman_Decomposition(M)
print("\nReverse Lu-Chipman Decomposition")
print("--------------------------------\n")
printMatrix("Diattenuator",diatten)
printMatrix("Retarder",ret)
printMatrix("Depolarizer",depol)
printMatrix("The product Diattenuator . Retarder .  Depolarizer", diatten @ ret @ depol)

diatten2, ret2, depol, ret1, diatten1 = Symmetric_Decomposition(M)
print("\nSymmetric decomposition")
print("-----------------------\n")
printMatrix("diatten2",diatten2)
printMatrix("ret2",ret2)
printMatrix("depol",depol)
printMatrix("ret1",ret1)
printMatrix("diatten1",diatten1)
printMatrix("The product Diattenuator2 . Retarder2 .  Depolarizer . Retarder1 . Diattenuator1",
            diatten2 @ ret2 @ depol @ ret1 @ diatten1)

L = MuellerLog(M)
print("\nLogarithmic decomposition")
print("-------------------------\n")
printMatrix("L",L)
printMatrix("MuellerExp(L)",MuellerExp(L))

print("\nSome other characteristic parameters")
print("------------------------------------\n")
print("Tmax = ",M.Tmax())
print("Tmin = ",M.Tmin())
print("diattenuation = ",M.diattenuation())
print("linear diattenuation = ",M.linear_diattenuation())
print("polarization dependent loss = ",M.polarization_dependent_loss())
print("polarizance = ",M.polarizance())
print("depolarization index = ",M.depolarization_index())
print("extinction_ratio = ",M.extinction_ratio())

parametersM = CharacterizedMueller(M)
print(parametersM)

print("==========================")
print("Testing pySCATMECH.fresnel")
print("==========================")

from pySCATMECH.fresnel import *

# We will be plotting some stuff, so we need matplotlib.pyplot
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker

wavelengths = np.arange(0.2,2.5,0.01)

# The following looks for a file named "silicon"
# Assume that "silicon.txt" exists on your computer...
silicon = OpticalFunction("silicon.txt") 

# A fixed real index:
glass = OpticalFunction(1.5) 

# A metal, i.e., a complex index of refraction:
aMetal = OpticalFunction(1.4+2.1j)

# A material with a Cauchy dispersion, given as a lambda function
# When providing a function, one must provide an iterable list of wavelengths because 
# the function is tabulated.
SiO2 = OpticalFunction(lambda L: 1.4580 + 0.00354/L**2, wavelengths)

# A couple materials with Sellmeier functions...
MgF = OpticalFunction(lambda L: math.sqrt(1+0.48755108*L**2/(L**2-0.04338408**2)
                                           +0.39875031*L**2/(L**2-0.09461442**2)
                                           +2.3120353*L**2/(L**2-23.793604**2)),
                                             wavelengths)

ZnO = OpticalFunction(lambda L: math.sqrt(1+1.347091*L**2/(L**2-0.062543**2)
                                           +2.117788*L**2/(L**2-0.166739**2)
                                           +9.452943*L**2/(L**2-24.320570**2)),
                                             wavelengths)

# It is useful to have one for vacuum:
vacuum = OpticalFunction(1)

wavelengths = np.arange(0.25,1.5,0.001)

# Plot the curves...
plt.figure()
plt.plot(wavelengths, [silicon(w).real for w in wavelengths], label="Si n")
plt.plot(wavelengths, [silicon(w).imag for w in wavelengths], label="Si k")
plt.plot(wavelengths, [glass(w) for w in wavelengths], label="glass")
plt.plot(wavelengths, [SiO2(w) for w in wavelengths], label="SiO2")
plt.plot(wavelengths, [MgF(w) for w in wavelengths], label="MgF")
plt.plot(wavelengths, [ZnO(w) for w in wavelengths], label="ZnO")
plt.legend()
plt.show()

# Create a film of ZnO (defined above) of thickness 0.5 micrometers...
film1 = Film(ZnO, thickness = 0.5)

# Create a quarter-wave thickness of ZnO for wavelength 0.5 micrometers...
film2 = Film(ZnO, waves = 1/4, wavelength = 0.5)

# create a quarter-wave thickness of ZnO for wavelength 0.5 micrometers for 45 degrees incidence...
film3 = Film(ZnO, waves = 1/4, wavelength = 0.5, angle = 45*deg)

print(film1)
print(film2)
print(film3)

H = Film(ZnO, waves = 1/4, wavelength = 0.5)
L = Film(MgF, waves = 1/4, wavelength = 0.5)

notch_filter = 5*[H,L] + [2*H] + 5*[L,H]

stack = FilmStack()

wavelength = 0.500 

# Return a Jones matrix
r = stack.reflectionCoefficient(60*deg, wavelength, vacuum, silicon)
print("Jones matrix = \n",r,"\n")

# Return a Mueller matrix
R = stack.R(60*deg, wavelength, vacuum, silicon)
print("Mueller matrix = \n",R,"\n")

# Return the s and p reflectance
Rs = stack.Rs(60*deg, wavelength, vacuum, silicon) 
Rp = stack.Rp(60*deg, wavelength, vacuum, silicon) 
print("Rs = ",Rs,", Rp = ",Rp)

plt.figure()
thetas=[t for t in np.arange(0,90)]
Rs = [stack.Rs(t*deg, wavelength, vacuum, silicon) for t in thetas]
Rp = [stack.Rp(t*deg, wavelength, vacuum, silicon) for t in thetas]
plt.plot(thetas, Rs, label="$R_\mathrm{s}$")
plt.plot(thetas, Rp, label="$R_\mathrm{p}$")
plt.legend()
plt.title("Reflectance of silicon at " + str(wavelength) + " µm")
plt.xlabel(r"$\theta$ (degrees)")
plt.ylabel("Reflectance")
plt.show()

plt.figure()
wavelengths = [L for L in np.arange(0.1, 1.5, 0.01)]
Rs = [stack.Rs(70*deg, wavelength, vacuum, silicon) for wavelength in wavelengths]
Rp = [stack.Rp(70*deg, wavelength, vacuum, silicon) for wavelength in wavelengths]
plt.plot(wavelengths, Rs, label="$R_\mathrm{s}(\lambda)$")
plt.plot(wavelengths, Rp, label="$R_\mathrm{p}(\lambda)$")
plt.legend()
plt.title("Reflectance of silicon at 70° incidence")
plt.xlabel("$\lambda$ (µm)")
plt.ylabel("Reflectance")
plt.show()

plt.figure()
wavelengths = [L for L in np.arange(0.1, 1.5, 0.01)]
M = [stack.R(70*deg, wavelength, vacuum, silicon) for wavelength in wavelengths]
delta = [(m @ Polarization(angle=45*deg)).delta() for m in M]
psi = [(m @  Polarization(angle=45*deg)).psi() for m in M]
plt.plot(wavelengths, delta, label="$\Delta(\lambda)$")
plt.plot(wavelengths, psi, label="$\Psi(\lambda)$")
plt.legend()
plt.title("Ellipsometric parameters")
plt.xlabel("$\lambda$ (µm)")
plt.ylabel('$\Psi$, $\Delta$ (rad)')
plt.show()

stack = FilmStack()
film = Film(SiO2, thickness = 0.01)
stack.grow(film)

plt.figure()
wavelengths = [L for L in np.arange(0.1, 1.5, 0.01)]

for t in [0., 0.01, 0.02, 0.03]:
    stack.clean()
    stack.grow(Film(SiO2, thickness = t))
    
    M = [stack.R(70*deg, wavelength, vacuum, silicon) for wavelength in wavelengths]
    delta = [(m @ StokesVector(1,0,1,0)).delta() for m in M]
    psi = [(m @ StokesVector(1,0,1,0)).psi() for m in M]
    plt.plot(wavelengths, delta)
    plt.plot(wavelengths, psi)

plt.title("Ellipsometric parameters")
plt.xlabel("$\lambda$ (µm)")
plt.ylabel('$\Psi$, $\Delta$ (rad)')
plt.show()

stack = FilmStack()

substrate = SiO2
wavelength = 0.500
stack.grow(Film(MgF, waves=1/4, wavelength=wavelength))

plt.figure()
ax = plt.subplot()
thetas=np.array([t for t in np.arange(0,90)])
Rs = [stack.Rs(theta*deg, wavelength, vacuum, substrate) for theta in thetas]
Rp = [stack.Rp(theta*deg, wavelength, vacuum, substrate) for theta in thetas]
Ts = [stack.Ts(theta*deg, wavelength, vacuum, substrate) for theta in thetas]
Tp = [stack.Tp(theta*deg, wavelength, vacuum, substrate) for theta in thetas]
plt.plot(thetas, Rs, label = "Rs")
plt.plot(thetas, Rp, label = "Rp")
plt.plot(thetas, Ts, label = "Ts")
plt.plot(thetas, Tp, label = "Tp")
plt.legend()
plt.xlim([-5, 95])
plt.xticks(np.linspace(0, 90, 10))
plt.xlabel("Angle")
plt.ylabel("Reflectance, Transmittance")
plt.title("Reflectance and Transmittance of a $\lambda/4$ MgF film on SiO$_2$")
class AddDegrees(ticker.Formatter):
    def __call__(self,x,pos):
        return "$%g \degree$"% x
ax.xaxis.set_major_formatter(AddDegrees())
class AddPercent(ticker.Formatter):
    def __call__(self,x,pos):
        return ("%g" % (x*100))+" %"
ax.yaxis.set_major_formatter(AddPercent())
plt.show()

L = Film(MgF, waves=1/4, wavelength=0.500)
H = Film(ZnO, waves=1/4, wavelength=0.500)

substrate = glass

def pltFilm(stack,label=""):
    stack = FilmStack(stack)
    wavelengths = np.arange(0.2, 1.0, 0.001)
    Rs = [stack.Rs(0*deg, L, vacuum, substrate) for L in wavelengths]
    plt.plot(wavelengths, Rs, label=label)

WORKING()
plt.figure()
plt.title("Buildup of a high reflector")
pltFilm(2*[L,H], label="2 LH")
pltFilm(3*[L,H], label="3 LH")
pltFilm(4*[L,H], label="4 LH")
pltFilm(10*[L,H], label="10 LH")
pltFilm(20*[L,H], label="20 LH")
plt.legend()
plt.xlabel("Wavelength (µm)")
plt.ylabel("Reflectance")
DONE()
plt.show()

WORKING()
plt.figure()
plt.title("Notch filter")
pltFilm(5*[H,L] + [2*H] + 5*[L,H],"5(H,L)+2H+5(L,H)")
plt.legend()
plt.xlabel("Wavelength (µm)")
plt.ylabel("Reflectance")
DONE()
plt.show()

print("========================")
print("Testing pySCATMECH.model")
print("========================")

from pySCATMECH.model import *

modelDictionary = getModelDictionary('Roughness_BRDF_Model')
printDictOfDict(modelDictionary)

print("Creating an ABC_PSD_Function")
psdFunction = Model("ABC_PSD_Function")
print("printParameters() returns:")
print(psdFunction.printParameters())

print("\ngetParameters() returns:")
print(psdFunction.getParameters())

print("\ngetParameterDictionary() returns:")
print(printDictOfDict(psdFunction.getParameterDictionary()))

psdFunction.setParameter('A',0.2)
psdFunction.getParameters()

parameters ={None : 'Bobbert_Vlieger_BRDF_Model',
              'lambda': '0.633',
              'substrate': '(1.56,0.2)',
              'type': '0',
              'density': '1',
              'sphere': '(1.59,0)',
              'radius': '1',
              'spherecoat': 'No_StackModel',
              'stack': 'SingleFilm_StackModel',
              'delta': '0',
              'quiet': 1} 
model = Model(parameters)
print(model)

printDictOfDict(model.getParameterDictionary())

model.getParameter('radius')

model.getModelName()



print("A dialog box should open. Choose a model, by double clicking on it.")
model = DialogGetModel('BRDF_Model')
print("""
A dialog box should open showing that model's parameters. Change parameters
as you like and then close window.""")
m = DialogGetModelParameters(Model(model))
print(model)

print("=======================")
print("Testing pySCATMECH.brdf")
print("=======================")

from pySCATMECH.brdf import *
import matplotlib.pyplot as plt

model = BRDF_Model('Correlated_Roughness_Stack_BRDF_Model')
print(model)

psd = {None : "ABC_PSD_Function",
       "A" : 0.01,
       "B" : 360,
       "C" : 2.4}

stack = {None : 'SingleFilm_StackModel',
         'material' : 1.59,
         'thickness': 0.05
        }

parameters = {'lambda' : 0.532,
              'substrate' : 4.05+0.05j,
              'type' : 0,
              'stack' : stack,
              'psd' : psd}

model = BRDF_Model('Correlated_Roughness_Stack_BRDF_Model',parameters)
print(model)

parameters['lambda'] = 0.600
psd['C'] = 2.3
stack['thickness'] = 0.1
model.setParameters(parameters)
print(model)

# Needed to use OpticalFunction, Film, and FilmStack...
from pySCATMECH.fresnel import *

# Define two materials...
SiO2 = OpticalFunction(lambda L: 1.4580 + 0.00354/L**2,np.arange(0.2,1.5,0.1))
Si = OpticalFunction('silicon')

# Define the power spectral density function...
psd = Model("ABC_PSD_Function", A=0.01, B=360, C=2.4)

# Define the stack...
stack =  FilmStack([Film(SiO2,thickness =0.05)])

model.setParameters(wavelength = 0.600, substrate=Si, stack = stack, psd = psd)
print(model)

model.getParameters()

print(model.MuellerBRDF(60*deg, 30*deg))

mBRDF = model.MuellerBRDF(60*deg, 30*deg)
inc = Polarization('p')
sens = Sensitivity('u')
print(sens @ mBRDF @ inc)

print(model.BRDF(60*deg, 30*deg, inc=Polarization('p'), sens=Sensitivity('u')))

print(model.BRDF(60*deg, 30*deg))

thetaslist = np.arange(-89, 90, 1)
BRDFlist = [model.BRDF(60*deg, thetas*deg, inc=[1,-1,0,0], sens=[1,0,0,0]) for thetas in thetaslist]

plt.figure()
plt.yscale('log')
plt.plot(thetaslist,BRDFlist)
plt.ylabel('BRDF (sr$^{-1}$)')
plt.xlabel(r'$\theta_\mathrm{r}$ (degrees)')
plt.xticks(np.linspace(-90,90,7))
plt.show()

thetaslist = np.arange(-89,90,1) 
plt.figure()
plt.yscale('log')
for thetai in [0,30,60]:
    BRDFlist = [model.BRDF(thetai*deg,thetas*deg,inc=[1,-1,0,0],sens=[1,0,0,0]) for thetas in thetaslist]
    plt.plot(thetaslist,BRDFlist,label = r"$\theta_i=%d^\circ$" % thetai)

plt.xticks(np.linspace(-90,90,7))
plt.legend()
plt.show()

thetaslist = np.linspace(-89,89,100) 

plt.figure()
for t in [0.1,0.2,0.3,0.4,0.5]:
    stack = FilmStack([Film(SiO2,thickness = t)])
    model.setParameters(stack=stack)
    BRDFlist = [model.BRDF(60*deg,thetas*deg,inc=[1,-1,0,0],sens=[1,0,0,0]) for thetas in thetaslist]
    plt.plot(thetaslist,BRDFlist,label = r"$D = %g$" % t)
    
plt.yscale('log')
plt.xticks(np.linspace(-90,90,7))
plt.legend()
plt.show()

print("========================")
print("Testing pySCATMECH.local")
print("========================")

import matplotlib.pyplot as plt
from pySCATMECH.local import *

spherecoat = {None : 'No_StackModel'}

stack = {None : 'No_StackModel'}

parameters = {'lambda' : 0.532,
              'substrate' : 4.06+0.05j,
              'type' : 0,
              'sphere' : 1.59,
              'radius' : 0.05,
              'spherecoat' : spherecoat,
              'stack' : stack,
              'delta' : 0,
              'lmax' : 0,
              'order' : -1,
              'Norm_Inc_Approx' : 0,
              'improve' : 3,
              'quiet' : 1}

model = Local_BRDF_Model("Bobbert_Vlieger_BRDF_Model",parameters)
print(model)

# Needed to use OpticalFunction, Film, and FilmStack...
from pySCATMECH.fresnel import *

# Define two materials...
PS = OpticalFunction(lambda L: (1+1.4435*L**2/(L**2-0.020216))**(1/2), np.arange(0.4356,1.052,0.01))
SiO2 = OpticalFunction(lambda L:(1+0.6961663*L**2/(L**2-0.0684043**2)+
                                   0.4079426*L**2/(L**2-0.1162414**2)+
                                   0.8974794*L**2/(L**2-9.896161**2))**(1/2), 
                                    np.exp(np.linspace(math.log(0.210),math.log(6.7),50)))

Si = OpticalFunction('silicon')

# Define the stack...
spherecoat =  FilmStack()
stack = FilmStack([Film(SiO2,thickness =0.0017)])

parameters = {'lambda' : 0.305,
              'substrate' : Si,
              'type' : 0,
              'sphere' : PS,
              'radius' : 0.05,
              'spherecoat' : spherecoat,
              'stack' : stack,
              'delta' : 0,
              'lmax' : 0,
              'order' : -1,
              'Norm_Inc_Approx' : 0,
              'improve' : 3}

model.setParameters(parameters)
print(model)

mDSC = model.MuellerDSC(60*deg, 30*deg)
print(mDSC)

thetaslist = np.linspace(-89, 89, 100)
DSClist = [model.DSC(60*deg, thetas*deg, inc=Polarization('p'), sens=Sensitivity('u')) for thetas in thetaslist]

plt.plot(thetaslist,DSClist)
plt.xlabel(r"$\theta_s$ [degrees]")
plt.ylabel(r"DSC [$\mathrm{\mu m}^2/\mathrm{sr}$]")
plt.xlim((-90,90))
plt.xticks(np.linspace(-90,90,7))
plt.show()

thetaslist = np.linspace(-89, 89, 100)

plt.figure()
for D in [0.1, 0.2, 0.5, 1.0]:
    parameters["radius"] = D/2
    model.setParameters(parameters)
    DSClist = [model.DSC(60*deg, thetas*deg, inc=Polarization('p'), sens=Sensitivity('u')) for thetas in thetaslist]
    plt.plot(thetaslist, DSClist, label = "D = %g µm" % D)

plt.xlabel(r"$\theta_s$ [degrees]")
plt.ylabel(r"DSC [$\mathrm{\mu m}^2/\mathrm{sr}$]")
plt.xlim((-90,90))
plt.yscale('log')
plt.xticks(np.linspace(-90,90,7))
plt.legend()
plt.show()


print("============================")
print("Testing pySCATMECH.integrate")
print("============================")

import matplotlib.pyplot as plt
from pySCATMECH.integrate import *

hemi = Hemisphere()

integrator1 = Integrator(2*deg, hemi, type=1)
integrator1.PlotSamplingPoints()

integrator2 = Integrator(2*deg, hemi, type=2)
integrator2.PlotSamplingPoints()

integrator3 = Integrator(2*deg, hemi, type=3)
integrator3.PlotSamplingPoints()

cone = CircularCone(theta=45*deg, phi=90*deg, alpha=20*deg)
    
integrator = Integrator(1*deg, cone)
integrator.PlotSamplingPoints()

cone = EllipticalCone(theta=0*deg, phi=0*deg, alpha=(20*deg, 45*deg))
    
integrator = Integrator(1*deg, cone)
integrator.PlotSamplingPoints()

cone = EllipticalCone(theta=0*deg, phi=0*deg, alpha=(20*deg, 45*deg), gamma=45*deg)
    
integrator = Integrator(1*deg, cone)
integrator.PlotSamplingPoints()


cone = RectangularCone(theta=0*deg, phi=0*deg, alpha=(20*deg, 45*deg))
    
integrator = Integrator(1*deg, cone)
integrator.PlotSamplingPoints()

cone = RectangularCone(theta=0*deg, phi=0*deg, alpha=(20*deg, 45*deg), gamma=45*deg)
    
integrator = Integrator(1*deg, cone)
integrator.PlotSamplingPoints()

cone = ProjectedPolygon([(0.5,0.5), 
                         (0.5,-0.5), 
                         (0,-0.2), 
                         (-0.5,-0.5), 
                         (-0.5,0.5), 
                         (0,0.2)])
    
integrator = Integrator(1*deg, cone)
integrator.PlotSamplingPoints()

hole1 = CircularCone(theta=70*deg, phi=0, alpha=5*deg)
hole2 = CircularCone(theta=70*deg, phi=180*deg, alpha=5*deg)

detector =  Hemisphere() & ~(hole1 | hole2)

integrator = Integrator(1*deg, detector)
integrator.PlotSamplingPoints()

circle = CircularCone(theta=0, phi=0, alpha=70*deg)
hole1 = CircularCone(theta=70*deg, phi=0, alpha=5*deg)
hole2 = CircularCone(theta=70*deg, phi=180*deg, alpha=5*deg)
hole3 = CircularCone(theta=0, phi=0, alpha=30*deg)
center = CircularCone(theta=0, phi=0, alpha=20*deg)

detector = circle & ~(hole1 | hole2 | hole3) | center

integrator = Integrator(1*deg, detector)
integrator.PlotSamplingPoints()

slot = RectangularCone(theta=0, phi=0, alpha=(90*deg, 6*deg))

detector3 = ~slot

integrator = Integrator(1*deg, detector3)
integrator.PlotSamplingPoints()

# The model parameters as a dictionary...
parameters = {'lambda' : 0.532,
              'substrate' : 4.05+0.05j,
              'type' : 0,
              'psd' : {None : 'Gaussian_PSD_Function',
                       'sigma' : 0.05,
                       'length' : 1
                       }
             }

# Define the model and set its parameters...
model = BRDF_Model("Microroughness_BRDF_Model", **parameters)

# Define the incident angle...
thetaIncident = 6*deg

# The detector is the hemisphere with a hole cut out of it to let in the 
# incident light and to let out the specular reflection...
detector = Hemisphere() & ~CircularCone(theta=0*deg, alpha=10*deg)

# Get the integration points and show them...
integrator = Integrator(1*deg, detector)
integrator.PlotSamplingPoints()

# Calculate and print the reflectance ...
ref = integrator.Reflectance(model, thetaIncident) 
print("Integrated reflectance = %g" % ref)

# Get the RMS roughnesses in logarithmic spacing...
sigmas = np.exp(np.linspace(np.log(0.0001), np.log(.01), 20)) 

# Initialize a list of reflectances...
refs = []

# s-polarized incident radiation...
spol = StokesVector(1, 1, 0, 0)

WORKING()
for sigma in sigmas:
    # Set the RMS roughness...
    model.setParameter("psd.sigma", sigma)
    
    # Calculate the reflectance...
    ref = integrator.Reflectance(model, thetaIncident, incpol=spol) 
    
    # Add it to the list..
    refs.append(ref)
DONE()

# Plot it...
plt.figure()
plt.loglog(sigmas,refs)
plt.xlabel("$\sigma$")
plt.ylabel("Reflectance")
plt.title("Reflectance versus RMS Roughness")
plt.show()

BVparameters = {'lambda' : 0.532,
              'substrate' : "silicon",
              'type' : 0, 
              'sphere' : 1.59,
              'radius' : 0.05,
              'spherecoat' : "No_StackModel",
              'stack' : {None : 'SingleFilm_StackModel',
                        'material' : '(1.5,0)',
                        'thickness' : '0.0016'},                         
                'delta' : 0,
                'quiet' : 1} # use defaults for operational parameters

BVmodel = Local_BRDF_Model("Bobbert_Vlieger_BRDF_Model",**BVparameters)

thetaIncident = 70*deg

circle = CircularCone(theta=0, phi=0, alpha=70*deg)
hole1 = CircularCone(theta=70*deg, phi=0, alpha=5*deg)
hole2 = CircularCone(theta=70*deg, phi=180*deg, alpha=5*deg)
hole3 = CircularCone(theta=0*deg, phi=0, alpha=30*deg)

detector = circle & ~(hole1 | hole2 | hole3)

integrator = Integrator(2*deg, detector)
integrator.PlotSamplingPoints()

cs = integrator.CrossSection(BVmodel,thetaIncident,incpol=[1,-1,0,0])

print("Cross section = %g µm^2" % cs)

diameters = np.exp(np.linspace(np.log(0.05),np.log(2),20)) #logarithmic spacing

refp = []
refs = []
spol = Polarization('s')
ppol = Polarization('p')
WORKING()
for d in diameters:
    BVmodel.setParameters(radius=d/2)
    refs.append( integrator.CrossSection(BVmodel, thetaIncident, incpol=spol) )
    refp.append( integrator.CrossSection(BVmodel, thetaIncident, incpol=ppol) )
DONE()
plt.figure()
plt.loglog(diameters,refs,label="s")
plt.loglog(diameters,refp,label="p")
plt.xlabel("Diameter / $\mathrm{\mu m}$")
plt.ylabel("Cross section / $\mathrm{\mu m}^2$")
plt.title("Cross Section versus Diameter")
plt.legend()
plt.show()

hemi = Hemisphere()
integrator = Integrator(1*deg, hemi)
dsc = integrator.DSC(BVmodel, thetaIncident, incpol=ppol)
integrator.PlotSamplingPoints(color=np.log(dsc), expand=3)

detector = CircularCone(theta=45*deg, phi=90*deg, alpha=30*deg, sensitivity=Sensitivity('s'))
integrator = Integrator(1*deg, detector)

diameters = np.exp(np.linspace(np.log(0.05), np.log(2),20)) #logarithmic spacing

WORKING()
ref = []
for d in diameters:
    BVmodel.setParameters(radius=d/2)
    ref.append( integrator.CrossSection(BVmodel, thetaIncident, incpol=Polarization('p')) )
DONE()

plt.figure()
plt.loglog(diameters, ref)
plt.xlabel("Diameter / $\mathrm{\mu m}$")
plt.ylabel("Cross section / $\mathrm{\mu m}^2$")
plt.title("Cross Section versus Diameter")
plt.show()

print("============================")
print("Testing pySCATMECH.scatterer")
print("============================")

import matplotlib.pyplot as plt
import numpy as np
from pySCATMECH.scatterer import *

parameters = {
            'lambda' : 0.532,
            'medium' : 1,
            'radius' : 0.1,
            'sphere' : 1.59}

model = Free_Space_Scatterer("MieScatterer",parameters)

print(model)

theta = 45*deg

vin = [0,0,1] 
vout = [np.sin(theta),0,np.cos(theta)]

m = model.DifferentialScatteringCrossSection(vin,vout)
print(m)

def dsc(theta):
    vin = [0,0,1]
    vout = [np.sin(theta),0,np.cos(theta)]
    return model.DifferentialScatteringCrossSection(vin,vout)
    
angles = np.linspace(0,2*pi,180)
results = [dsc(angle) for angle in angles]

spol = Polarization('S') # perpendicular to the scattering plane
ppol = Polarization('P') # parallel to the scattering plane
upol = Polarization('U') # unpolarized 
usens = Sensitivity('U') # unpolarized sensitivity

plt.figure()
plt.polar(angles, [usens @ r @ upol for r in results], label = "unpol")
plt.polar(angles, [usens @ r @ spol for r in results], label = "perp")
plt.polar(angles, [usens @ r @ ppol for r in results], label = "par")
plt.legend()
plt.show()

parameters = {
            'lambda' : 0.532,
            'medium' : 1,
            'radius' : 0.03,
            'sphere' : 'gold'}

model = Free_Space_Scatterer("MieScatterer", parameters)

def integrate_scatter(model, wavelength, steps):
    "Integrate `model` at `wavelength`, performing the integral in `steps` steps."
    integ = 0
    dtheta = pi / (steps - 1)
    parameters['lambda'] = wavelength
    model.setParameters(parameters)
    vin = [0, 0, 1]
    angles = np.arange(0, pi, dtheta)
    # Integrate angles from 0 to pi...
    for theta in angles:
        vout = [np.sin(theta), 0, np.cos(theta)]
        m = model.DifferentialScatteringCrossSection(vin, vout)
        integ += m[0][0] * np.abs(np.sin(theta))
    integ *= 2 * pi * dtheta
    return integ

wavelengths = np.linspace(0.380,0.780,200)

plt.figure()

# Diameter 0.04 µm...
D = 0.04
parameters['radius'] = D/2
model.setParameters(parameters)

crosssection = [integrate_scatter(model,L,50) for L in wavelengths]

plt.plot(wavelengths, crosssection, label = "D = %g µm"%D)

plt.legend(frameon=False)
plt.show()

parameters = {
            'lambda' : 0.532,
            'medium' : 'water', 
            'radius' : 0.03,
            'sphere' : 'glass',
            'coating' : 'gold',
            'thickness' : 0.005}

model = Free_Space_Scatterer("CoatedMieScatterer", parameters)

def extinction(model, wavelength):
    "Calculate the extinction coefficient at a given `wavelength` for a given `model`."
    parameters['lambda'] = wavelength
    model.setParameters(parameters)
    vin = [0, 0, 1]
    return model.Extinction(vin)[0][0]

def integrated_scatter(model,wavelength,steps):
    "Integrate the scatter over polar angle"
    integ = 0
    dtheta = pi/(steps - 1)
    parameters['lambda'] = wavelength
    model.setParameters(parameters)
    vin = [0, 0, 1]
    for theta in np.linspace(0, pi, steps):
        vout = [np.sin(theta), 0, np.cos(theta)]
        m = model.DifferentialScatteringCrossSection(vin, vout)
        integ += m[0][0]*np.abs(np.sin(theta))
    integ *= 2*pi*dtheta
    return integ

wavelengths = np.linspace(0.500, 1.300, 200)

fig = plt.figure(figsize=(10,5))
ax1 = fig.add_subplot(131)
ax2 = fig.add_subplot(132)
ax3 = fig.add_subplot(133)

WORKING()

for t in [0.002,0.003,0.004,0.005,0.006,0.007,0.008]:
    parameters['thickness'] = t
    sca = np.array([integrated_scatter(model,L,50) for L in wavelengths])
    ext = np.array([extinction(model,L) for L in wavelengths])
    abs = ext-sca
    ax1.plot(wavelengths, sca, label = "t = %g"%t)
    ax2.plot(wavelengths, ext, label = "t = %g"%t)
    ax3.plot(wavelengths, abs, label = "t = %g"%t)
DONE()

ax1.legend()
ax1.set_title("Scattering")
ax2.set_title("Extinction")
ax3.set_title("Absorption")
plt.show()

parameters = {
            'lambda' : 0.589,
            'medium' : 1,
            'radius' : 1,
            'materialright' : 1.56+ 6.9E-7,
            'materialleft' : 1.56}

model = Free_Space_Scatterer("Optically_Active_Sphere_Scatterer",parameters)

def extinction(model,wavelength):
    model.setParameters(wavelength=wavelength)
    vin = [0,0,1]
    return model.Extinction(vin)

extmatrix = model.Extinction([0,0,1])

print(extmatrix)

shape = {None : 'Ellipsoid_Axisymmetric_Shape',
        'npoints' : '100',
        'vertical' : '0.010',
        'horizontal' : '0.003'
        }

tmatrix = {'lambda' : 0.8,
        'medium' : 1.4,
        'Shape' : shape,
        'particle' : 'gold',
        'lmax' : '30',
        'mmax' : '0'}

model = Free_Space_Scatterer("TMatrix_Axisymmetric_Scatterer",tmatrix)

extCoeff100 = model.Extinction([1,0,0])
extCoeff001 = model.Extinction([0,0,1])

print("Mueller matrix transmittance perpendicular to ellipsoid axis:")
print(MuellerExp(-extCoeff100 * 10**5),'\n')
print("Mueller matrix transmittance parallel to ellipsoid axis:")
print(MuellerExp(-extCoeff001 * 10**5),'\n')

print("======================")
print("Testing pySCATMECH.rcw")
print("======================")


import matplotlib.pyplot as plt
import numpy as np
from pySCATMECH.rcw import *


gratingParameters = {
    None : 'Single_Line_Grating', # The Grating model can be assigned with the `None` key
    'period' : 0.2,               # The period  
    'medium_i' : 1,               # The incident medium0
    'medium_t' : 'silicon',       # The transmitting (substrate) material
    'material' : 'silicon',       # The grating material
    'space' : 1,                  # The space material
    'height' : 0.2,               # The height
    'topwidth' : 0.1,             # Top width
    'bottomwidth' : 0.15,         # Bottom width
    'offset' : 0,                 # Offset of top, relative to bottom
    'nlevels' : 10}               # Number of layers in the staircase approximation

rcwParameters = {
    'order' : 25,                 # number of Floquet orders considered
    'type' : 0,                   # 0 indicates reflection
    'lambda' : 0.532,             # wavelength
    'thetai' : 70,                # incident angle in degrees 
    'rotation' : 0,               # rotation angle in degrees
    'grating' : gratingParameters,
    'quiet' : 1
    }

model = RCW_Model(rcwParameters)

print(model)

# 250 wavelengths from 0.2 to 1.5...
wavelengths = np.linspace(0.2, 1.5, 250)


def rcw(wavelength):
    rcwParameters['lambda'] = wavelength
    model.setParameters(rcwParameters)
    return model.DiffractionEfficiency(0)

WORKING()
muellerMatrices = [rcw(wavelength) for wavelength in wavelengths]
DONE()

N = [m[0,1] for m in muellerMatrices]
C = [m[2,2] for m in muellerMatrices]
S = [m[2,3] for m in muellerMatrices]

plt.plot(wavelengths*1000, N, label="N")
plt.plot(wavelengths*1000, C, label="C")
plt.plot(wavelengths*1000, S, label="S")
plt.xlabel('$\lambda$ / nm')
plt.legend()
plt.show()

wavelengths = np.linspace(0.2, 1.5, 75)

rcwParameters['rotation'] = 45
model.setParameters(rcwParameters)


def rcw(wavelength):
    rcwParameters['lambda'] = wavelength
    model.setParameters(rcwParameters)
    return model.DiffractionEfficiency(0)

WORKING()
muellerMatrices = [rcw(wavelength) for wavelength in wavelengths]
DONE()

plt.figure(figsize = [13,13])
for element in range(16):
    plt.subplot(4,4,element+1)
    if element==0:
        plt.ylim(0,1)
    else:
        plt.ylim(-1,1)
    plt.plot(wavelengths,[m[element//4,element%4] for m in muellerMatrices],'b-')

plt.show()

def showGrating(model,x,z):
    xx, zz = np.meshgrid(x, z, sparse=True)

    def eps(x,z):
        return model.getEpsilon(x,z).real

    epsilon = np.array([[eps(x0,z0) for x0 in x] for z0 in z])

    fig = plt.figure(figsize=(6,5))
    ax = fig.add_subplot(111)
    h = ax.contourf(x,z,epsilon,256,cmap="brg")
    plt.colorbar(h,ax=ax)

    plt.show()
    
showGrating(model,np.linspace(-0.2, 0.2, 200),np.linspace(-0.3,.1,200))

gratingParameters = {
    None : 'Single_Line_Grating',
    'period' : 10,
    'medium_i' : 1,
    'medium_t' : 'silicon',  
    'material' : 'silicon',
    'space' : 1,
    'height' : '0.5',
    'topwidth' : '0.4',
    'bottomwidth' : '0.15',
    'offset' : '0',
    'nlevels' : '10'}

rcwParameters = {
    None : 'RCW_Model',
    'order' : '25',
    'type' : '0',
    'lambda' : '0.532',
    'thetai' : 45,
    'rotation' : 45,
    'grating' : gratingParameters,
    'quiet' : 1
    }

model.setParameters(rcwParameters)

def plotDiffractionPattern(model):
    WORKING()
    order = int(model.getParameter("order"))
    x, y, I = [], [], []
    for i in range(-order,order):
        direction = model.Direction(i)
        if direction[2] != 0 :
            x.append(direction[0])
            y.append(direction[1])
            I.append(model.DiffractionEfficiency(i)[0,0])
       
    thetas = np.linspace(0, 2*pi, 100)
    circlex = np.cos(thetas)
    circley = np.sin(thetas)
    
    plt.figure(figsize = (10,10))
    ax = plt.subplot(111)
    ax.scatter(x, y, s=np.array(I)*100) 
    ax.plot(circlex, circley, c = 'r')
    DONE()
    plt.show()

plotDiffractionPattern(model)

# Set the lines per millimeter:
lpm = 600
# Set the nominal wavelength in µm:
blazeLambda = 1.00

##########################
period = 1000/lpm
sinBlaze = blazeLambda/2/period
height = period*sinBlaze
##########################

# Description of the grating...
gratingParameters = {None: 'Triangular_Grating', 
                     'period': 1000/lpm, 
                     'medium_i': 1, 
                     'medium_t': 'aluminum', 
                     'material': 'aluminum',
                     'amplitude' : height,
                     'aspect' : 1,
                     'nlevels' : 20}

# The RCW parameters
rcwParameters = {
    None : 'RCW_Model',
    'order' : 25,
    'type' : 0,
    'lambda' : 0.500,
    'thetai' : 0,
    'rotation' : 0,
    'grating' : gratingParameters,
    'quiet' : 1
    }
model = RCW_Model(rcwParameters)

def efficiency(wavelength):
    """Set the wavelength and calculate the efficiency"""
    thetai=math.asin(wavelength/2/period)
    model.setParameters(wavelength=wavelength,thetai=thetai)
    return model.DiffractionEfficiency(1)

wavelengths = np.arange(0.2, 1.5, 0.02)

WORKING()
eff = [efficiency(L) for L in wavelengths]
DONE()

effs = [Sensitivity('u') @ e @ Polarization('s') for e in eff]
effp = [Sensitivity('u') @ e @ Polarization('p') for e in eff]

plt.figure()
plt.plot(wavelengths, effs, label = "s-pol")
plt.plot(wavelengths, effp, label = "p-pol")
plt.legend()
plt.show()

showGrating(model,np.linspace(-6, 6, 200),np.linspace(-6,6,200))

# Set the lines per millimeter:
lpm = 1000
# Sinusoid amplitude:
amplitude = .5
# Incident angle:
thetai = 45*deg


period = 1000/lpm
from pySCATMECH.fresnel import OpticalFunction

FusedSilica = OpticalFunction(lambda L: 1.4580 + 0.00354/L**2, np.arange(0.2,1.5,0.05))

# Description of the grating...
gratingParameters = {None: 'Sinusoidal_Relief_Grating', 
                     'period': period, 
                     'medium_i': 1, 
                     'medium_t': FusedSilica, 
                     'material': FusedSilica,
                     'amplitude' : amplitude,
                     'base' : 0,
                     'option' : 0,
                     'nlevels' : 20}

# The RCW parameters
rcwParameters = {
    None : 'RCW_Model',
    'order' : 10,
    'type' : 1,
    'lambda' : 0.500,
    'thetai' : 0,
    'rotation' : 0,
    'grating' : gratingParameters,
    'quiet' : 1
    }
model = RCW_Model(rcwParameters)

def efficiency(wavelength):
    """Set the wavelength and calculate the efficiency"""
    rcwParameters['lambda'] = wavelength
    model.setParameters(rcwParameters)
    return model.DiffractionEfficiency(1)

wavelengths = np.arange(0.2, 1.5, 0.02)

WORKING()
eff = [efficiency(L) for L in wavelengths]
DONE()

effs = [Sensitivity('u') @ e @ Polarization('s') for e in eff]
effp = [Sensitivity('u') @ e @ Polarization('p') for e in eff]

plt.figure()
plt.plot(wavelengths, effs, label = "s-pol")
plt.plot(wavelengths, effp, label = "p-pol")
plt.legend()
plt.show()

showGrating(model,np.linspace(-6, 6, 200),np.linspace(-6,6,200))

print("===========================")
print("Testing pySCATMECH.crossrcw")
print("===========================")


import matplotlib.pyplot as plt
import numpy as np
from pySCATMECH.crossrcw import *

period = 0.5
order = 5
theta = 70

gratingParameters = {
    None : 'Cylinder_CrossGrating',
              'medium_i' : 1,
              'medium_t' : "fusedsilica",
              'zeta' : 0,
              'd1' : period,
              'd2' : period,
              'grid1' : 256,
              'grid2' : 256,
              'rtop' : 0.1,
              'rbottom' : 0.1,
              'thickness' : 0.1,
              'nlevels' : 1,
              'inside' : 1,
              'outside' : "fusedsilica"}

rcwParameters = {
    'thetai' : theta,
    'rotation' : 0,
    'lambda' : 0.532,
    'type' : '0',
    'order1' : order,
    'order2' : order,
    'grating' : gratingParameters
    }

model = CrossRCW_Model(rcwParameters)
print(model)

wavelengths = np.linspace(0.2, 1.5, 100)

def rcw(wavelength):
    model.setParameter('lambda',wavelength)
    return model.DiffractionEfficiency(0,0)

WORKING()
muellerMatrices = [rcw(wavelength) for wavelength in wavelengths]
DONE()

N = [m[0,1] for m in muellerMatrices]
C = [m[2,2] for m in muellerMatrices]
S = [m[2,3] for m in muellerMatrices]

plt.figure()
plt.plot(wavelengths*1000, N, label="N")
plt.plot(wavelengths*1000, C, label="C")
plt.plot(wavelengths*1000, S, label="S")
plt.xlabel('$\lambda$/nm')
plt.legend()
plt.show()

um2eV = 1.239841775 
energies = np.linspace(um2eV/0.2,um2eV/1.5,10)

rcwParameters['rotation'] = 22.5

def rcwA(energy):
    rcwParameters["lambda"] = um2eV/energy
    model.setParameters(rcwParameters)
    m = model.DiffractionEfficiency(0,0)
    m, m[0,0] = m/m[0,0],m[0,0] # Normalize the matrix, keeping the m[0,0] element
    return m

WORKING()
muellerMatrices = [rcwA(E) for E in energies]
DONE()

plt.figure(figsize = [13,13])
for element in range(16):
    plt.subplot(4,4,element+1)
    plt.ylim(-1,1)
    plt.plot(energies,[muellerMatrices[i][element//4,element%4] for i in range(len(energies))],'b:')

plt.show()

period = 2
order = 10
theta = 25

gratingParameters = {
    None : 'Cylinder_CrossGrating',
              'medium_i' : 1, 
              'medium_t' : 1, # Free standing
              'zeta' : 30,    # Hexagonal pattern
              'd1' : period,
              'd2' : period,
              'grid1' : 256,
              'grid2' : 256,
              'rtop' : period/4,
              'rbottom' : period/4,
              'thickness' : 100, # Very deep
              'nlevels' : 1,
              'inside' : 1,     # Holes in
              'outside' : 1.5   # glass
            }

rcwParameters = {
    'thetai' : theta,
    'rotation' : 0,
    'lambda' : 0.532,
    'type' :   1,  # in transmission
    'order1' : order,
    'order2' : order,
    'grating' : gratingParameters
    }

model.setParameters(rcwParameters)

def plotDiffractionPattern(model):
    WORKING()
    order1 = int(model.getParameter("order1"))
    order2 = int(model.getParameter("order2"))
    x, y, I = [], [], []
    for i in range(-order1,order1):
        for j in range(-order2,order2):
            # Direction() function gets the direction 
            direction = model.Direction(i,j) 
            if direction[2] != 0 :
                x.append(direction[0])
                y.append(direction[1])
                I.append(model.DiffractionEfficiency(i,j)[0,0])
       
    thetas = np.linspace(0, 2*pi, 100)
    circlex = np.cos(thetas)
    circley = np.sin(thetas)

    plt.figure(figsize = (10,10))
    ax = plt.subplot(111)
    ax.scatter(x, y, s=np.array(I)*2000) # You may need to adjust this size scaling
    ax.plot(circlex, circley, c = 'r')
    DONE()
    plt.show()

plotDiffractionPattern(model)

import os
# The following file got created by TMatrix_Axisymmetric_Scatterer and should be removed...
os.remove('shape.dat')

print("========")
print("Finished")
print("========")



import SCATPY
from pySCATMECH.model import *
from pySCATMECH.mueller import *

class RCW_Model(Model):
    """
    Class handling rigorous coupled wave (RCW) analysis for 1D gratings.
    """
    def __init__(self, *args, **kwargs):
        """
        args and kwargs are forwarded to Model.setParameters() and contain
        model parameters
        """
        self.newer = SCATPY.Get_RCW_Model
        self.freer = SCATPY.Free_RCW_Model
        self.handle = SCATPY.Get_RCW_Model()
        self.setParameters(*args, **kwargs)
        
    def __del__(self):
        SCATPY.Free_RCW_Model(self.handle)

    def DiffractionEfficiency(self, i):
        """
        Returns the Mueller matrix diffraction efficiency for 
        the i-th diffraction order.

        Parameters
        ----------

        i : int
            Diffraction order

        Returns
        -------

        efficiency : MuellerMatrix
                     The Mueller matrix diffraction efficiency for order i

        """
        return MuellerMatrix(SCATPY.RCWDiffractionEfficiency(self.handle,i))

    def DiffractionAmplitude(self, i):
        """
        Returns the Jones matrix diffraction amplitude for 
        the i-th diffraction order.

        Parameters
        ----------

        i : int
            Diffraction order

        Returns
        -------

        amplitude : JonesMatrix
                     The Jones matrix diffraction amplitude for order i

        """
        return np.array(SCATPY.RCWDiffractionAmplitude(self.handle,i))

    def Direction(self, i):
        """
        Returns a 3D unit vector [x,y,z] in the direction of 
        propagation of the i-th diffraction order
        Parameters
        ----------

        i : int
            Diffraction order

        Returns
        -------
        direction : list of float
                    Directional cosines of diffracted order

        """
        return SCATPY.RCWDirection(self.handle, i)

    def getEpsilon(self, x, z, direction='x'):
        """
        Returns the grating dielectric constant at a specific location
        
        Parameters
        ----------

        x, z : float
               Location coordinates 

        direction : str, optional
                    If the grating is anisotropic, then direction can be set
                    to choose a particular field direction. Can be 'x', 'y', or 
                    'z'. (Default is 'x').

        Returns
        -------

        epsilon : complex
                  The dielectric constant evuated at the current wavelength 
                  and grating.
        """
        if direction=='x': d=0
        elif direction=='y': d=1
        elif direction=='z': d=2
        else: d=0

        return SCATPY.GetGratingEpsilon(self.handle,x,z,d)


    def getGratingDefinition(self):
        """
        Returns the grating definition.


        Returns
        -------

        Dictionary containing the following keys:

        "position"    : list of lists of x-positions of the transitions for 
                        each of the layers in the grating.
        "materialx"   : list of lists of the dielectric constants (xx element)
                        in each of the layers of the grating.
        "materialy"   : list of lists of the dielectric constants (yy element) 
                        in each of the layers of the grating.
        "materialz"   : list of lists of the dielectric constants (zz element) 
                        in each of the layers of the grating.
        "materialmux" : list of lists of the magnetic susceptibility (xx 
                        element) in each of the layers of the grating.
        "materialmuy" : list of lists of the magnetic susceptibility (yy 
                        element) in each of the layers of the grating.
        "materialmuz" : list of lists of the magnetic susceptibility (zz 
                        element) in each of the layers of the grating.
        "thickness"   : list of thicknesses of each layer of the grating.
        "heights"     : list of the heights of each interface of the grating.

        If the grating is a Generic_Grating, then an additional key "segments" 
        contains a list of line segments for each of the boundaries, each 
        element a dict containing keys "x1", "y1", "x2", "y2", "mat1", and
        "mat2".

        """

        definition =  SCATPY.GetGratingDefinition(self.handle)

        definition['heights'] = [0]
        accum = 0
        for thickness in definition['thickness']:
            accum = accum - thickness
            definition['heights'].append(accum)

        return definition
            
    
